/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import DeviceLib
import Foundation
import os

/* The Logger
 *
 * An example of using the log command line tool to stream the
 * running application log output.
 *
 * $ sudo log stream --level default --predicate \
 *  'subsystem == "com.oracle.iot.sample" && category == "GatewayDeviceSample"'
 */
fileprivate let logger:OSLog = OSLog(subsystem: "com.oracle.iot.sample",
                                     category: "BeaconGatewaySample")

fileprivate let IBEACON_SENSOR_MODEL_URN:String =
    "urn:com:oracle:iot:device:location:ibeacon"
fileprivate let EDDYSTONE_SENSOR_MODEL_URN:String =
    "urn:com:oracle:iot:device:location:eddystone-tlm-uid"

// Set this to true to create "non roaming" indirectly connected devices
fileprivate let RESTRICTED:Bool = false

// Common beacon device model attributes
fileprivate let BEACON_RSSI_ATTRIBUTE:String = "ora_rssi"
fileprivate let BEACON_TXPOWER_ATTRIBUTE:String = "ora_txPower"
fileprivate let EDDYSTONE_TEMP_ATTRIBUTE:String = "temperature"
fileprivate let EDDYSTONE_VOLT_ATTRIBUTE:String = "batteryVoltage"

fileprivate let NUMBER_OF_SIMULATED_BEACONS:Int = 2

/**
 * Only information from beacons currently within this maximum distance in
 * meters will be reported.
 */
fileprivate let MAX_DISTANCE:Double = 100.0

func getMaxDistance() -> Double {
    return MAX_DISTANCE
}

public class BeaconGatewaySample : BeaconGatewaySampleBase {
    // Use real beacons or simulated beacons
    var simulate:Bool = false
    let beaconScanner:BeaconScanner?
    /**
     * The BeaconGatewaySample constructor
     */
    public init(path:String, password:String, updateInterval:Int,
                simulate:Bool = false,
                errorCallback: @escaping (String)->()) throws {
        self.simulate = simulate
        self.beaconScanner = self.simulate ? nil : BeaconScanner()
        
        try super.init(path:path, password:password,
            deviceModels: [
                IBEACON_SENSOR_MODEL_URN,
                EDDYSTONE_SENSOR_MODEL_URN
            ],
            updateInterval: updateInterval,
            errorCallback: { msg in
                os_log("%@", log:logger, type:.default, msg)
                errorCallback(msg)
            })
    }
    
    /**
     * Subclesses implement this method.
     */
    override
    func processSensorData(device:Beacon) {
        let rssi:Int = device.getRssi()
        let distance:Double = device.getDistance()
        guard let _ = device.getVirtualizedBeacon(),
                rssi != 0, distance <= MAX_DISTANCE else {
            return
        }
        switch (device.getType()) {
        case .IBEACON:
            updateIBeacon(beacon: device as! IBeacon)
            break
        case .EDDYSTONE, .EDDYSTONE_TLM:
            updateEddystoneBeacon(beacon: device as! EddystoneBeacon)
            break
        case .NONE:
            break
        }
    }
    
    func updateIBeacon(beacon:IBeacon) {
        guard let vbeacon:VirtualDevice = beacon.getVirtualizedBeacon() else {
            return
        }
        do {
            _ = try vbeacon.set(attributeName: BEACON_RSSI_ATTRIBUTE,
                            attributeValue: beacon.getRssi() as AnyObject)
            let output = String(format:"%@ : Set : \"ora_rssi\" = %d",
                             formatDate(date:Date()), beacon.getRssi())
            beacon.setOutput(output: output)
            os_log("%@", log:logger, type:.default, output)
        } catch {
            self.errorCallback("\(error)")
        }
    }
    
    // Return rssi from the virtual device
    func getIBeaconData(beacon:IBeacon) -> Int {
        var rssi:Int = 0
        guard let vbeacon:VirtualDevice = beacon.getVirtualizedBeacon() else {
            let msg:String = String(format:"No VirtualDevice for beacon %@",
                                    beacon.getHardwareId())
            errorCallback(msg)
            return rssi
        }
        do {
            rssi = try vbeacon.get(attributeName: BEACON_RSSI_ATTRIBUTE) as! Int
        } catch {
            errorCallback("\(error)")
        }
        return rssi
    }

    func updateEddystoneBeacon(beacon:EddystoneBeacon) {
        guard let vbeacon = beacon.getVirtualizedBeacon() else {
            return
        }
        let temp:Double = beacon.getTemperature()
        let voltage:Double = beacon.getVoltage()
        do {
            _ = try vbeacon.update()
            .set(attributeName: BEACON_RSSI_ATTRIBUTE,
                 attributeValue: beacon.getRssi() as AnyObject)
            .set(attributeName: EDDYSTONE_TEMP_ATTRIBUTE,
                 attributeValue: temp as AnyObject)
            .set(attributeName: EDDYSTONE_VOLT_ATTRIBUTE,
                 attributeValue: voltage as AnyObject)
            .finish()
            let output:String = String(format:
                "%@ : Set : \"ora_rssi\" = %d, \"temeperature\" = %.2f, \"batteryVoltage\" = %.2f",
                formatDate(date:Date()), beacon.getRssi(),
                ((temp / 256) * 100).rounded()/100, (voltage * 100).rounded()/100)
            beacon.setOutput(output:output)
            os_log("%@", log:logger, type:.default, output)
        } catch {
            self.errorCallback("\(error)")
        }
    }

    // Return the eddystone data, as tuple (rssi,temperature,voltage)
    func getEddystoneData(beacon:EddystoneBeacon) -> (rssi:Int?,temperature:Double?,voltage:Double?) {
        var data:(rssi:Int?,temperature:Double?,voltage:Double?) = (nil,nil,nil)
        guard let vbeacon:VirtualDevice = beacon.getVirtualizedBeacon() else {
            let msg:String = String(format:"No VirtualDevice for beacon %@",
                                    beacon.getHardwareId())
            errorCallback(msg)
            return data
        }

        do {
            data.rssi = try vbeacon.get(attributeName: BEACON_RSSI_ATTRIBUTE) as? Int
            data.temperature = try vbeacon.get(attributeName: EDDYSTONE_TEMP_ATTRIBUTE) as? Double
            data.voltage = try vbeacon.get(attributeName: EDDYSTONE_VOLT_ATTRIBUTE) as? Double
        } catch {
            errorCallback("\(error)")
        }
        return data
    }
    
    /*
     * This method is called after the gateway endpointId is established.
     * It is used as a key to create and persist the hardware id of a sensor
     * which is used to register the device.
     *
     * Called with endpointId non-nil during gateway initialization.
     */
    override
    func getDevices(endpointId:String?=nil) -> [Beacon] {
        var beacons:[Beacon] = [Beacon]()
        if let endpointId = endpointId {
            if simulate {
                // Define the hardware id on simulated devices
                // we need the gateway's endpoint id. Ideally the
                // scanner would just return a list of beacons simulated
                // or real, but there is a convention used by JCL for
                // ICD endpoint ids for simulated beacons.
                var beacon:Beacon
                for i in 0..<NUMBER_OF_SIMULATED_BEACONS {
                    if i % 2 == 0 {
                        beacon = IBeacon.getSimulatedInstance(minorNum: UInt8(i),
                            beaconUUID: "\(endpointId)_Sample")
                        beacons.append(beacon)
                    beacon.start()
                    } else {
                        beacon = EddystoneBeacon.getSimulatedInstance(minorNum: UInt8(i),
                            beaconUUID: "\(endpointId)_Sample")
                        beacons.append(beacon)
                        beacon.start()
                    }
                }
                beacons.sort(by: Beacon.sort)
                return beacons
            } else {
                return beaconScanner?.getBeacons() ?? beacons
            }
        }
        return deviceQueue.sync(execute: {
            for (_,device) in devices {
                beacons.append(device)
            }
            return beacons
        })
    }
    
    /*
     * Intialize VirtualDevice default values and setup control handlers
     */
    override
    func configureVirtualDevice(device:Beacon) {
        let mssi = device.getMssi()
        do {
            if let vbeacon:VirtualDevice = device.getVirtualizedBeacon() {
                _ = try vbeacon.set(attributeName: BEACON_TXPOWER_ATTRIBUTE,
                                    attributeValue: mssi as AnyObject)
            }
        } catch {
            self.errorCallback("\(error)")
        }
    }
    
    // Call this to determine if there are any virtual devices configured
    func haveBeacons() -> Bool {
        return deviceQueue.sync(execute: {
            return self.devices.count > 0
        })
    }
}

public class BeaconGatewaySampleBase {
    
    var gateway:GatewayDevice?
    
    /* The callbacks implemented by the View to render errors
     */
    var errorCallback:(String)->()?
    
    /* Update the virtual devices at a rate defined by this `updateInterval` in
     * seconds
     */
    let updateInterval:Double!
    /* The schedule timer that periodically updates the virtual devices at
     * the rate of `updateInterval`
     */
    var updateTimer:Timer!
    
    
    /* The supported device models */
    let deviceModels:[String]!
    
    /* This dispatch queue synchronizes access to the devices array */
    let deviceQueue:DispatchQueue = DispatchQueue(label:"Device Queue",
                                                  qos:.utility)
    /* Managed devices that are ready to run */
    var devices:[String:Beacon] = [String:Beacon]()
    func addDevice(device:Beacon) {
        deviceQueue.sync() {
            devices[device.getEndpointID()!] = device
        }
    }
    
    /* Make sure displays don't overwrite each other */
    let displayQueue:DispatchQueue = DispatchQueue(label:"Display Queue",
                                                   qos:.utility)
    
    // The GatewayDeviceSample interface requires a sensor update interval
    // and callbacks to update the View
    public init(path:String, password:String, deviceModels:[String],
                updateInterval:Int,
                errorCallback: @escaping (String)->()) throws {
        self.deviceModels = deviceModels
        self.errorCallback = errorCallback
        self.updateInterval = Double(updateInterval)
        
        gateway = try GatewayDevice(path: path, password: password)
        
        DispatchQueue.main.async(execute: {
            self.updateTimer = Timer.scheduledTimer(timeInterval: self.updateInterval,
                target: self,
                selector: #selector(self.updateVirtualDevices),
                userInfo: nil, repeats: true)
        })
    }
    
    /**
     * Subclesses implement this method.
     */
    func processSensorData(device:Beacon) {
        fatalError("Subclass must implement 'processSensorData'")
    }
    
    /**
     * Subclasses implement this method to configure the virtual devices
     * with virtual device control handlers and set default values in the
     * sensors.
     */
    func configureVirtualDevice(device:Beacon) {
        fatalError("Subclass must implement 'configureVirtualDevice`")
    }
    
    /**
     * Subclasses implement this method and return the set of devices that
     * it manages.
     */
    func getDevices(endpointId:String?=nil) -> [Beacon]? {
        fatalError("Subclass must implement 'getDevices'")
    }
    
    /*
     * Create a virtualDevice.
     * The virtualDevice instance is recorded in device.
     */
    func createVirtualDevice(device:Beacon) {
        do {
            try self.gateway!.getDeviceModel(deviceModelUrn: device.getDeviceModel(),
                                             callback: {(deviceModel,error) in
                if let error = error {
                    self.errorCallback("\(error)")
                    return
                }
                do {
                    let virtualDevice:VirtualDevice =
                        try self.gateway!.createVirtualDevice(
                            deviceId: device.getEndpointID()!,
                            deviceModel: deviceModel!)
                    
                    device.setVirtualizedBeacon(virtualizedBeacon: virtualDevice)
                    
                    self.configureVirtualDevice(device:device)
                    // The device is ready to run
                    self.addDevice(device:device)
                } catch {
                    self.errorCallback("\(error)")
                }
            })
        } catch {
            errorCallback("\(error)")
        }
    }
    
    /*
     * This method creates and initializes local sensors using "getDevices()" and
     * registers those devices. If the icd's were previously registered
     * they will not be recreated in the cloud.
     */
    func registerDevices(endpointId:String, deviceModels:[String]) {
        // TODO: If not simulating devices, need to do this until
        // TODO: devices are found, and periodically if new devices
        guard let beacons:[Beacon] = self.getDevices(endpointId:endpointId),
                beacons.count > 0 else {
            self.errorCallback("registerDevices(): No beacons found")
            return
        }
        /*
         * Once the Gateway endpoint id is obtained, the managed devices are
         * created and then registered.
         * When they are registered their endpoint ids are recorded.
         * Then virtual devices are created for each one.
         */
        for device in beacons {
            do {
                try self.gateway!.registerDevice(restricted: device.isRestricted(),
                                                 hardwareId: device.getHardwareId(),
                                                 metaData: device.getMetadata(),
                                                 deviceModels: device.getDeviceModel(),
                                                 callback: { endpointId, error in
                    if let error = error {
                        self.errorCallback("\(error)")
                        return
                    }
                    // Record the icd's endpointId
                    device.setEndpointID(endpointID: endpointId!)
                    self.createVirtualDevice(device:device)
                })
            } catch {
                self.errorCallback("\(error)")
            }
        }
    }
    
    /*
     * Activate the gatewy
     * Create and initialize sensors
     * Register sensors and create virtual devices for each one
     * Start the update loop.
     */
    public func start() {
        /*
         * First the Gateway must be activated. The endpointId is used
         * to uniquely identify icd's. Once the endpointId is obtained
         * local sensors can be created and registered.
         * If the gateway is already activated, then the endpointId is
         * is known, and previously registered icd's with the gateway
         * can be recreated without creating new icd's in the cloud.
         */
        do {
            if !self.gateway!.isActivated() {
                try self.gateway!.activate(callback: {endpointId, error in
                    if error != nil {
                        self.errorCallback("\(error!)")
                        return
                    }
                    // the gateway has this endpointId set on it during
                    // the activate call
                    
                    // For each device create a task that registers the device, and
                    // creates a virtual device for it.
                    self.registerDevices(endpointId:endpointId!,
                                         deviceModels: self.deviceModels)
                })
            } else {
                self.registerDevices(endpointId:self.gateway!.getEndpointId(),
                                     deviceModels:deviceModels)
            }
        } catch {
            errorCallback("\(error)")
        }
    }
    
    // Update the humidity and temperature sensor virtual devices attributes.
    @objc func updateVirtualDevices() {
        self.deviceQueue.sync() {
            for (_,device) in self.devices {
                processSensorData(device:device)
            }
        }
    }
    
    public func stop() {
        updateTimer.invalidate()
        for beacon in getDevices() ?? [Beacon]() {
            beacon.stop()
        }
        deviceQueue.sync() {
            self.devices.removeAll()
        }
        do {
            try close()
        } catch {
            let endpointId:String = gateway?.getEndpointId() ?? ""
            let msg = "stop(): \(endpointId) close() threw exception \(error)"
            os_log("%@", log:logger, type:.default, msg)
        }
    }
    
    public func close() throws {
        try gateway?.close()
        gateway = nil
    }
    
    deinit {
        do {
            try gateway?.close()
        } catch {
            os_log("deinit(): Caught exception closing BeaconGateway %@",
                   log: logger, type:.default, "\(error)")
        }
        gateway = nil
    }
}

