/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit


// Define custom table view cell
class DeviceSelectionCell : UITableViewCell {
    @objc static let ID = "DeviceCell"
    @IBOutlet weak var beaconInfo: UILabel!
    weak var beacon:Beacon!
}

// NOTE: To use simulated beacons, set the SIMULATE_BEACONS to true
fileprivate let SIMULATE_BEACONS:Bool = true //false

class BeaconSelectionView: UIViewController, UITableViewDelegate, UITableViewDataSource {

    @IBOutlet var SecondView: UIView!
    @IBOutlet weak var selectBeaconLabelOutlet: UILabel!
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var stopButtonOutlet: UIButton!
    
    var beaconGateway:BeaconGatewaySample?
    var tableupdatetimer:Timer?
    
    override func viewDidLoad() {
        super.viewDidLoad()
        // Do any additional setup after loading the view.
        self.tableView.delegate = self
        self.tableView.dataSource = self
        self.tableView.register(UITableViewCell.self,
                                forCellReuseIdentifier: "cell")
        
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.addObserver(self, selector: #selector(self.terminate),
                                               name: UIApplication.willTerminateNotification, object: nil)
        self.title = NAVBARTITLE
        do {
            try beaconGateway = BeaconGatewaySample(
                path:Provisioning.getProvisioningFilePath(),
                password:Provisioning.getProvisioningPassword(),
                updateInterval:5,
                simulate:SIMULATE_BEACONS,
                errorCallback:errorCallback)
            
            beaconGateway?.start()
        } catch {
            showError(controller: self, alertTitle: "BeaconGatewaySample startup failure",
                      message: "\(error)", actionTitle: "OK",
                      handler: { (action: UIAlertAction?) -> () in
                        // Ensure that user can re-provision the device
                        removeLastSettingsFile()
                        self.navigationController?.popToRootViewController(animated: false)
                        // TODO: Should this initiatiate the "terminate" sequence ?
            })
        }
    }
    
    // TODO: Remove references to SecondView and define all UI
    // TODO: attributes in UI builder
    override func viewWillAppear (_ animated: Bool) {
        super.viewWillAppear(animated)
        if tableupdatetimer == nil {
            DispatchQueue.main.async() {
                self.tableupdatetimer = Timer.scheduledTimer(timeInterval: 5.0, target: self,
                                              selector: #selector(BeaconSelectionView.reloadTable),
                                              userInfo: nil, repeats: true)
            }
        }
        if Provisioning.isProvisioningStatic() {
            self.navigationItem.setHidesBackButton(true, animated: false)
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        // back button clicked (Reset)
        if isMovingFromParent {
            Provisioning.reProvision()
            navigationController?.popToRootViewController(animated: false)
        }
    }
    
    override func viewDidAppear (_ animated: Bool) {
        super.viewDidAppear(animated)
        self.tableView.flashScrollIndicators()
    }
    
    var beaconList:[Beacon]? = nil
    func tableView(_ tableView: UITableView,
                   numberOfRowsInSection section: Int) -> Int {
        var count = 0
        guard let beaconGateway = self.beaconGateway else {
            return count
        }
        beaconList = beaconGateway.getDevices()
        for b in beaconList ?? [Beacon]() {
            if b.getDistance() <= getMaxDistance() {
                count += 1
            }
            else {
                break
            }
        }
        return count
    }
    
    // Fill the table view with device models.
    func tableView(_ tableView: UITableView,
                   cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        
        let cell: DeviceSelectionCell = self.tableView.dequeueReusableCell(
            withIdentifier: DeviceSelectionCell.ID) as! DeviceSelectionCell
        
        guard let beaconList:[Beacon] = self.beaconList else {
            return cell
        }
        cell.beaconInfo.textAlignment = NSTextAlignment.left

        guard beaconList.count > 0 else {
            return cell
        }
        let hardId = beaconList[indexPath.row].getHardwareId()
        let distance = beaconList[indexPath.row].getDistance()
        if distance != -1 {
            cell.beaconInfo.text = hardId +
                String(format: " \n  %.3f meters away", distance)
        } else {
            cell.beaconInfo.text = hardId + " \n  Distance unknown"
        }
        cell.beaconInfo.adjustsFontSizeToFitWidth = true
        cell.accessoryType = UITableViewCell.AccessoryType.disclosureIndicator
        cell.beacon = beaconList[indexPath.row]
        
        return cell
    }
    
    func tableView(_ tableView: UITableView,
                   didSelectRowAt indexPath: IndexPath) {
        // Can this ever be nil ?
        if let cell:DeviceSelectionCell = self.tableView.cellForRow(at: indexPath) as? DeviceSelectionCell {
            cell.selectionStyle = .gray
            self.showBeacon(beacon: cell.beacon)
        } else {
            showError(controller: self, alertTitle: "BeaconGatewaySample Selection Error",
                      message: "Beacon selection returned nil", actionTitle: "OK",
                      handler: { (action: UIAlertAction?) -> () in  })
        }
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    private func errorCallback(str:String) {
        showError(controller: self, alertTitle: "Library Error",
                  message: str, actionTitle: "OK",
                  handler: { (action: UIAlertAction?) -> () in })
    }
    
    // Action for one row selection. Active devices are got according to the 
    // selected device model.
    func showBeacon(beacon: Beacon) {
        performSegue(withIdentifier: "devicesegue", sender: beacon)
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        let dest:BeaconGatewaySampleView = segue.destination as! BeaconGatewaySampleView
        // Assume we do have beacons since it was checked before performing the
        // segue
        dest.selectedBeacon = sender as? Beacon
        dest.selectedBeacon?.start()
        dest.beaconGateway = beaconGateway
    }
    
    @objc func reloadTable() {
        if beaconGateway?.haveBeacons() ?? false {
            self.tableView.reloadData()
        }
    }
    
    // Action for the button "Stop"
    @IBAction func stopButtonAction(_ sender: UIButton) {
        tableupdatetimer?.invalidate()
        tableupdatetimer = nil
        // When the app is suspended, because of the Info.plist property
        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
        // observer is called and the "terminate" method will be called.
        UIControl().sendAction(#selector(NSXPCConnection.suspend),
                               to: UIApplication.shared, for: nil)
    }
    
    // Close the data connector and all used resources,
    // gateway device, tam and communication channel
    @objc func terminate() {
        tableupdatetimer?.invalidate()
        tableupdatetimer = nil
        NotificationCenter.default.removeObserver(self)
        beaconGateway?.stop()
        saveSettingsFile()
    }
}



