/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit

var dcdSample:DirectlyConnectedDeviceSample?

@objc class DirectlyConnectedDeviceSampleView: UIViewController, UITextViewDelegate {
    
    // The virtual device update interval in seconds
    let UPDATE_INTERVAL:Int = 5
    
    //MARK: Properties
    @IBOutlet weak var humidityIdOutlet: UILabel!
    @IBOutlet weak var humidityOutlet: UILabel!
    @IBOutlet weak var humidityThresholdOutlet: UILabel!
    
    @IBOutlet weak var stopButtonOutlet: UIButton!
    
    @IBOutlet weak var resetButtonOutlet: UIButton! {
        didSet {
            if Provisioning.isProvisioningStatic() {
                resetButtonOutlet.isHidden = true
            }
        }
    }
    
    @IBOutlet weak var outputLabelOutlet: UILabel!
    @IBOutlet weak var outputTextViewOutlet: UITextView!
    @IBOutlet var ViewControllerOutlet: UIView!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        outputTextViewOutlet.delegate = self
        self.title = NAVBARTITLE
        
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.addObserver(self, selector: #selector(self.terminate),
                                               name: UIApplication.willTerminateNotification, object: nil)
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        do {
            try dcdSample = DirectlyConnectedDeviceSample(
                path:Provisioning.getProvisioningFilePath(),
                password:Provisioning.getProvisioningPassword(),
                updateInterval:UPDATE_INTERVAL,
                usePolicy:Provisioning.getWithPolicies(),
                appCallback:appCallback)
            
            dcdSample?.start()
        
        } catch {
            showError(controller: self, alertTitle: "DirectlyConnectedDeviceSample startup failure",
                      message: "\(error)", actionTitle: "OK",
                      handler: { (action: UIAlertAction?) -> () in
                // Ensure that user can re-provision the device
                removeLastSettingsFile()
                self.navigationController?.popToRootViewController(animated: false)
            })
        }
        if Provisioning.isProvisioningStatic() {
            self.navigationItem.setHidesBackButton(true, animated: false)
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        // back button clicked (Reset)
        if isMovingFromParent {
            Provisioning.reProvision()
            navigationController?.popToRootViewController(animated: false)
        }
    }

    private func appCallback(text:String?, sensorData:[String:Any]?, str:String?) {
        if let text = text {
            display(text:text)
        }
        if let sensorData = sensorData {
            updateGui(sensorData:sensorData)
        }
        if let str = str {
            errorCallback(str:str)
        }
    }

    private func display(text:String) {
        DispatchQueue.main.async(execute: {
            self.outputTextViewOutlet.text =
                self.outputTextViewOutlet.text + "\n" + text
            // Force text field to (invisibly) scroll until latest line is at bottom
            let bottom: NSRange = NSMakeRange(self.outputTextViewOutlet.text.count - 10, 1)
            self.outputTextViewOutlet.scrollRangeToVisible((bottom))
        })
    }
    
    private func errorCallback(str:String) {
        showError(controller: self, alertTitle: "Library Error",
                  message: str, actionTitle: "OK",
                  handler: { (action: UIAlertAction?) -> () in })
    }

    /*
     * We know there is only one HumiditySensor and one TemperatureSensor
     * The humidity data is obtained with key "humidity" as (String, Int, Int)
     * (endpointId,humidity,maxThreshold)
     * and temperature data is obtained with key "temperature"
     * as (String, Int64, Double, Double, Double, Int, Int)
     * (endpointId,startTime,temperature,minTemp,maxTemp,minThreshold,maxThreshold)
     */
    func updateGui(sensorData:[String:Any]) {
        
        // TODO: This should be done with discrete labels, not as a single
        // string in one label. This will allow updating the humidity or
        // temperature sensor independently

        if let anyData = sensorData["humidity"] {
            
            let hData:(endpointId:String,humidity:Int,maxThreshold:Int) =
                anyData as! (String,Int,Int)
            
            humidityIdOutlet.text = hData.endpointId
            humidityOutlet.text = hData.humidity.description + " %"
            humidityThresholdOutlet.text = hData.maxThreshold.description + " %"
        }
    }

    // MARK: Actions
    
    // Action for the button "Reset"
    @IBAction func onResetButton(_ sender: UIButton) {
        dcdSample?.stop()
        // stop the loops
        outputTextViewOutlet.text = ""
        // Allow reprovisioning
        Provisioning.reProvision()
        self.navigationController?.popToRootViewController(animated: false)  
    }
    
    // Action for the button "Stop"
    @IBAction func onStopButton(_ sender: UIButton) {
        outputTextViewOutlet.text = ""
        // When the app is suspended, because of the Info.plist property
        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
        // observer is called and the "terminate" method will be called.
        UIControl().sendAction(#selector(NSXPCConnection.suspend),
            to: UIApplication.shared, for: nil)
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // Close directly connected device, and exit
    @objc func terminate() {
        NotificationCenter.default.removeObserver(self)
        dcdSample?.stop()
        saveSettingsFile()
    }
}
