/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// The DeviceMonitorView is common to all device monitor views. It
// encapsulates common properties for instances of EnterpriseClientSample,
// the selected Device and VirtualDevice

class BeaconDeviceMonitorView: DeviceMonitorView {

    @IBOutlet var BeaconDeviceMonitorViewOutlet: UIView!
    @IBOutlet weak var deviceNameLabelOutlet: UILabel!
    @IBOutlet weak var deviceIdLabelOutlet: UILabel!
    @IBOutlet weak var manufacturerLabelOutlet: UILabel!
    @IBOutlet weak var modelNumberLabelOutlet: UILabel!
    @IBOutlet weak var serialNumberLabelOutlet: UILabel!
    @IBOutlet weak var deviceUUIDLabelOutlet: UILabel!
    @IBOutlet weak var beaconMajorLabelOutlet: UILabel!
    @IBOutlet weak var beaconMinorLabelOutlet: UILabel!
    @IBOutlet weak var measuredSignalLabelOutlet: UILabel!
    @IBOutlet weak var receivedSignalLabelOutlet: UILabel!
    
    let rssiValueFormat:String = "Received: %.1f"
    let ora_txPowerValueFormat:String = "Measured: %d"
    let uuidValueFormat:String = "UUID: %@"
    let majorValueFormat:String = "Major: %@"
    let minorValueFormat:String = "Minor: %@"

    override func viewDidLoad() {
        super.viewDidLoad()
        self.title = NAVBARTITLE

        // Monitor the selected device for iBeacon device model and update the 
        // UI view accordingly.
        guard let virtualDevice = virtualDevice else {
            return
        }
            
        // iBeacon values changes are managed.
        virtualDevice.setOnChange(callback: { (event: ChangeEvent) in
            DispatchQueue.main.async(execute: {
                var namedValue = event.getNamedValue()
                repeat  {
                    switch namedValue.getName() {
                    case "ora_rssi":
                        let rssi:Double = (namedValue.getValue() as! NSNumber).doubleValue
                            self.receivedSignalLabelOutlet.text =
                                String(format:self.rssiValueFormat, rssi)
                        break
                    case "ora_txPower":
                        let mssi:Int = (namedValue.getValue() as! NSNumber).intValue
                        self.measuredSignalLabelOutlet.text =
                            String(format:self.ora_txPowerValueFormat, mssi)
                        break
                    default:
                        break
                    }
                    guard let next:NamedValue = namedValue.next() else {
                        break
                    }
                    namedValue = next
                }
                while true
            })
        })
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    override func viewWillAppear (_ animated: Bool) {
        deviceIdLabelOutlet.text = self.virtualDevice?.getEndpointId()
        
        manufacturerLabelOutlet.text = "Manufacturer: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MANUFACTURER.rawValue)) ?? "")
        modelNumberLabelOutlet.text = "Model Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MODEL_NUMBER.rawValue)) ?? "")
        serialNumberLabelOutlet.text = "Serial Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? "")
        
        let metadata:[String:String]? = device?.getMetadata()
        self.deviceUUIDLabelOutlet.text = "UUID: " + (metadata?["UUID"] ?? "Unknown")
        self.beaconMajorLabelOutlet.text = "Major: " + (metadata?["minor"] ?? "Unknown")
        self.beaconMinorLabelOutlet.text = "Minor: " + (metadata?["major"] ?? "Unknown")
        
        do {
            let power:Int =
                (try virtualDevice?.get(attributeName: "ora_txPower") as? NSNumber)?.intValue ?? 0
            self.measuredSignalLabelOutlet.text = String(format:ora_txPowerValueFormat, power)
            let rssi:Double =
                (try virtualDevice?.get(attributeName: "ora_rssi") as? NSNumber)?.doubleValue ?? 0.0
            self.receivedSignalLabelOutlet.text = String(format:rssiValueFormat, rssi)
        }
        catch {
            print(error)
        }
    }
}
