/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib


// Select a DeviceModel to see its active devices.
class DeviceModelSelectionView: UIViewController, UITableViewDelegate,
                                                  UITableViewDataSource {

    // the EnterpriseClientSample instance
    var enterpriseClientSample:EnterpriseClientSample? = nil
    // the application context
    var applicationName:String? = nil
    // The supported device models
    var deviceModels:[String:DeviceModel]? = nil
    // A convenient array to work with the table view
    var deviceModelsArray:[(name:String,model:DeviceModel)] =
        [(name:String,model:DeviceModel)]()
    
    var timer:Timer?
    // Check for device model updates
    var updateInterval:TimeInterval = 1.0

    @IBOutlet var deviceSelectionViewOutlet: UIView!
    @IBOutlet var exitApp: UIButton!

    @IBOutlet weak var deviceModelsLabelOutlet: UILabel!
    @IBOutlet weak var tableView: UITableView!

    // Create the EnterpriseClientSample instance.
    // Put this in a method since it can happen in ViewOnLoad or
    // 'viewWillAppear'. If the back button is clicked, we want "start over"
    // so the "enterpriseClientSample" will be set to 'nil'. If it is
    // 'nil' in 'viewWillAppear' then recreate it.
    //
    // This class can be used to perfom the library tasks necessary
    // to obtain the data populate the various device monitoring screens
    // and this DeviceModelSelection screen.
    //
    // Because nework calls are made EnterpriseClientSample returns immediately.
    // The instance may not be ready to use. Set a flag to wait for the callback
    // to be invoked. The instance is ready to be used. This is a common pattern
    // when using library apis as can be seen later in the file.
    func intializeEnterpriseClientSample() {
        self.enterpriseClientSample = EnterpriseClientSample(callback: { error in
            if let error = error {
                // TODO: show alert
                print(error)
                self.terminate()
            }
        },
        applicationName:applicationName)
    }
    
    @objc func updateDeviceModels() {
        guard let ecsample = self.enterpriseClientSample else {
            print("EnterpriseClientSample is not set")
            return
        }
        guard let deviceModels:[String:DeviceModel] = ecsample.getDeviceModels() else {
            print("DeviceModelSelectView.updateDeviceModels(): no device models")
            return
        }
        // TODO: be more efficient
        deviceModelsArray.removeAll()
        self.tableView.reloadData()
        for (name,model) in deviceModels {
            deviceModelsArray.append((name,model))
        }
        self.tableView.reloadData()
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        self.tableView.delegate = self
        self.tableView.dataSource = self
        self.tableView.register(UITableViewCell.self,
                                forCellReuseIdentifier: "cell")
        
        intializeEnterpriseClientSample()
        self.title = NAVBARTITLE

    }

    override func viewWillAppear (_ animated: Bool) {
        if Provisioning.isProvisioningStatic() {
            self.navigationItem.setHidesBackButton(true, animated: false)
        }
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.addObserver(self, selector: #selector(self.terminate),
                                               name: UIApplication.willTerminateNotification, object: nil)
        if enterpriseClientSample == nil {
            intializeEnterpriseClientSample()
        }
        super.viewWillAppear(false)
        DispatchQueue.main.async() {
            self.timer = Timer.scheduledTimer(timeInterval: self.updateInterval,
                                              target: self,
                                              selector: #selector(self.updateDeviceModels),
                                              userInfo: nil, repeats: true)
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(false)
        DispatchQueue.main.async() {
            self.timer?.invalidate()
        }
        
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.removeObserver(self)
        
        // back button clicked (Reset)
        if isMovingFromParent {
            Provisioning.reProvision()
            navigationController?.popToRootViewController(animated: false)
        }
    }
    
    override func viewDidAppear (_ animated: Bool) {
        super.viewDidAppear(false)
        self.tableView.flashScrollIndicators()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        // The number of device models
        return self.deviceModelsArray.count
    }

    // Fill the table view with device model names.
    func tableView(_ tableView: UITableView,
                   cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell:UITableViewCell = self.tableView.dequeueReusableCell(withIdentifier: "cell")!
        cell.textLabel?.textAlignment = NSTextAlignment.left
        cell.textLabel!.text = self.deviceModelsArray[indexPath.row].name
        cell.accessoryType = .disclosureIndicator
        return cell
    }

    // A DeviceModel was selected
    // Display the DeviceSelectionScreen to load the devices.
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let cell:UITableViewCell? = self.tableView.cellForRow(at: indexPath)
        cell?.selectionStyle = .gray
        displayDeviceSelectionView(deviceModel:deviceModelsArray[indexPath.row].model)
    }
    
    func displayDeviceSelectionView(deviceModel:DeviceModel) {
        // If we are here we know that enterpriseClientSample is not nil
        // and that devices is not nil
        performSegue(withIdentifier: "DevModelSelToDevSel",
                     sender: (self.enterpriseClientSample!, deviceModel))
    }
    
    // sender is (<EnterpriseClientSample>,<DeviceModel>)
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        let sender = sender as! (EnterpriseClientSample,DeviceModel)
        let enterpriseClientSample:EnterpriseClientSample = sender.0
        let deviceModel:DeviceModel = sender.1
        let dest:DeviceSelectionView = segue.destination as! DeviceSelectionView
        dest.enterpriseClientSample = enterpriseClientSample
        dest.deviceModel = deviceModel
    }
    
    func displayServerLoginView() {
        let serverLoginView = self.storyboard?.instantiateViewController(
            withIdentifier: "ServerLoginView") as! ServerLoginView
        self.navigationController?.pushViewController(serverLoginView,
                                                      animated: false)
    }
    
    func handleError(error:String, fatal:Bool=true) {
        if isUserAuthenticationError(error:error) {
            self.enterpriseClientSample = nil
            displayServerLoginView()
        } else {
            print(error)
            // TODO: Exit ?
            showError(controller: self, alertTitle: "Enterprise Client Alert",
                      message: "\(error)", actionTitle: "OK",
                      handler: {(action:UIAlertAction?) -> () in
                        // When the app is suspended, because of the Info.plist property
                        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
                        // observer is called and the "terminate" method will be called.
                        if fatal {
                            UIControl().sendAction(#selector(NSXPCConnection.suspend),
                                               to: UIApplication.shared, for: nil)
                        }
            })
        }
    }
    // actions

    @IBAction func backButtonOutletAction(_ sender: UIButton) {
        // If leaving the DeviceModelSelection screen then we are returning to the
        // provisioning screen or the application screen.
        // TODO: Will viewOnLoad be called again ?
        self.enterpriseClientSample?.close(callback: { error in
            // TODO: call "showError"
            if let error = error {
                print("Error while closing enterprise client \(error)")
            }
            self.enterpriseClientSample = nil
            self.deviceModelsArray.removeAll()
        })
        dismiss(animated: false, completion: nil)
    }
    
    // TODO: defer this to the application selection screen and use the
    // NAV back button to get there. Although this might be problematic
    // if the device model selection view does not come from the
    // application selection view.
    @IBAction func onExitAppClick(_ sender: UIButton) {
        // When the app is suspended, because of the Info.plist property
        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
        // observer is called and the "terminate" method will be called.
        UIControl().sendAction(#selector(NSXPCConnection.suspend),
                               to: UIApplication.shared, for: nil)
    }
    
    // Close directly connected device, and exit
    @objc func terminate() {
        self.enterpriseClientSample?.close(callback: { error in
            if let error = error {
                print(error)
            }
        })
        saveSettingsFile()
    }
}

