/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// Define custom table view cell
class DeviceSelectionCell : UITableViewCell {
    @objc static let ID = "DeviceCell"
    @IBOutlet weak var deviceId: UILabel!
    var device:Device? = nil
    var virtualDevice:VirtualDevice? = nil
}

// Convenience class to encapsulate these common variables.
// All the *MonitorDeviceView classes extend this class
class DeviceMonitorView : UIViewController {
    var enterpriseClientSample:EnterpriseClientSample? = nil
    var virtualDevice: VirtualDevice? = nil
    var device:Device? = nil
    required init?(coder:NSCoder) {
        super.init(coder:coder)
    }
}

// This view displays the supported device models
class DeviceSelectionView: UIViewController, UITableViewDelegate, UITableViewDataSource {
    
    @IBOutlet var deviceSelectionViewOutlet: UIView!
    @IBOutlet weak var tableView: UITableView!
    
    var enterpriseClientSample:EnterpriseClientSample? = nil
    var deviceModel:DeviceModel? = nil
    var devices:[(Device,VirtualDevice)]? = nil
    
    let DEVICEIDTEXT:String = "Device ID"
    var BOLDDEVICEIDTEXT = NSMutableAttributedString()
    
    let noDeviceViewData:(sensor:String,label:String,view:String,type:Any.Type) = ("", "No Sensor", "", Any.self)
    var deviceViewData:[String:(sensor:String,label:String,view:String,type:Any.Type)] = [
        HUMIDITY_SENSOR_DEVICE_MODEL_URN:("Humidity", "Humidity Sensors",
             "HumidityDeviceMonitorView", HumidityDeviceMonitorView.Type.self),
        TEMPERATURE_SENSOR_DEVICE_MODEL_URN:("Temperature", "Temperature Sensors",
             "TemperatureDeviceMonitorView", TemperatureDeviceMonitorView.Type.self),
        IBEACON_DEVICE_MODEL_URN:("RSSI", "iBeacon Sensors",
             "BeaconDeviceMonitorView", BeaconDeviceMonitorView.Type.self),
        EDDYSTONEBEACONTLMUID_DEVICE_MODEL_URN:("RSSI", "Eddystone Sensors",
             "EddystoneDeviceMonitorView", EddystoneDeviceMonitorView.Type.self),
        MOTIONACTIVATEDCAMERA_DEVICE_MODEL_URN:("Motion Activated Camera", "Motion Activated Cameras",
             "MotionActivatedCameraView", MotionActivatedCameraView.Type.self)
    ]

    override func viewDidLoad() {
        // The segue to this view should have set the DeviceModel
        guard let deviceModel = self.deviceModel else {
            handleError(error:"DeviceSelectionView.viewDidLoad(): " +
                "Required deviceModel instance has not been set.", fatal:false)
            return
        }
        // Get the active devices for the selected device model.
        // Errors are not treated as fatal. Show an alert pop up
        // and dismiss back to the DeviceModelSelection screen.
        var _error:String? = nil
        var callbackReturned:Bool = false
        self.enterpriseClientSample?.getDevices(deviceModel: deviceModel,
                                          callback: { devices, error in
            _error = error
            self.devices = devices
            callbackReturned = true
        })
        while callbackReturned == false {
            RunLoop.current.run(until: Date(timeIntervalSinceNow: 1.0))
        }
        if let error = _error {
            handleError(error:error, fatal:false)
            return
        }
        guard let _ = self.devices else {
            handleError(error:"DeviceModelSelectionView.tableView(didSelectRowAt): " +
                "getDevices() returned nil error and devices list for device model " +
                "\(self.deviceModel?.getName() ?? "Unknown")", fatal:false)
            return
        }
        guard let _devices = self.devices, _devices.count > 0 else {
            print("no devices")
            showError(controller: self, alertTitle: "Select Device Model",
                message: "No \(self.deviceModel?.getName() ?? "unknown device model") devices", actionTitle: "OK",
                handler:nil)
            self.navigationController?.popViewController(animated: false)
            return
        }
        super.viewDidLoad()
        
        self.tableView.delegate = self
        self.tableView.dataSource = self
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "cell")
        // Long press action implemented for each table row to see better entire
        // long device ids.
        let holdToShowDeviceId = UILongPressGestureRecognizer(target: self,
                                                              action: #selector(DeviceSelectionView.longShowId))
        holdToShowDeviceId.minimumPressDuration = 1.10
        self.tableView.addGestureRecognizer(holdToShowDeviceId)
        
        self.title = NAVBARTITLE
        
        // Make the Header cell centered and bold
        let aligncenter = NSMutableParagraphStyle()
        aligncenter.alignment = .center
        let attrs = [
            NSAttributedString.Key.font : UIFont.boldSystemFont(ofSize: 18),
            NSAttributedString.Key.paragraphStyle:aligncenter
        ]
        let bolddeviceid = NSMutableAttributedString(string: DEVICEIDTEXT,
                                                         attributes: attrs)
        BOLDDEVICEIDTEXT.append(bolddeviceid)
    }
    
    override func viewWillAppear (_ animated: Bool) {
        super.viewWillAppear(animated)
        //self.tableView.reloadData()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
    }
    
    override func viewDidAppear (_ animated: Bool) {
        super.viewDidAppear(animated)
        self.tableView.flashScrollIndicators()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return (self.devices?.count ?? 0) + 1
    }
    
    func tableView(_ tableView: UITableView,
                   cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        var index:Int = indexPath.row
        let cell: DeviceSelectionCell = self.tableView.dequeueReusableCell(
            withIdentifier: DeviceSelectionCell.ID) as! DeviceSelectionCell
        cell.deviceId.textAlignment = .left
        
        // Header Row
        if index == 0 {
            cell.deviceId.attributedText = BOLDDEVICEIDTEXT
            cell.accessoryType = .none
            return cell
        }
        index = index - 1
        guard let device:(Device,VirtualDevice) = self.devices?[index] else {
            handleError(error:"DeviceSelectionView.tableView(cellForRowAt): " +
                "No device for index \(index)", fatal:false, dismiss:false)
            return cell
        }
        cell.device = device.0
        cell.virtualDevice = device.1
        cell.deviceId.text = cell.device?.getId()
        cell.accessoryType = .disclosureIndicator
        
        return cell
    }
    
    // Show monitor details for selected active device.
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let index:Int = indexPath.row - 1
        guard index >= 0,
                let cell:DeviceSelectionCell =
                    self.tableView.cellForRow(at: indexPath) as? DeviceSelectionCell,
                let device:Device = cell.device,
                let virtualDevice:VirtualDevice = cell.virtualDevice else {
            print("DeviceSelectionView.tableView(didSelectRowAt): " +
                  "do not have a Device and VirtualDevice for row \(indexPath.row)")
            return
        }
        cell.selectionStyle = .gray
        self.monitorDevice(device:device, virtualDevice:virtualDevice)
    }
    
    func displayServerLoginView() {
        let serverLoginView = self.storyboard?.instantiateViewController(
            withIdentifier: "ServerLoginView") as! ServerLoginView
        self.navigationController?.pushViewController(serverLoginView,
                                                      animated: false)
    }
    
    func handleError(error:String, fatal:Bool=true, dismiss:Bool=true) {
        if isUserAuthenticationError(error:error) {
            self.enterpriseClientSample = nil
            displayServerLoginView()
        } else {
            print(error)
            // TODO: Exit ?
            showError(controller: self, alertTitle: "Enterprise Client Alert",
                      message: "\(error)", actionTitle: "OK",
                      handler: {(action:UIAlertAction?) -> () in
                        // When the app is suspended, because of the Info.plist property
                        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
                        // observer is called and the "terminate" method will be called.
                        if fatal {
                            UIControl().sendAction(#selector(NSXPCConnection.suspend),
                                                   to: UIApplication.shared, for: nil)
                        } else if dismiss {
                            self.dismiss(animated: false, completion: nil)
                        }
            })
        }
    }
    
    // actions

    func monitorDevice(device:Device, virtualDevice:VirtualDevice) {
        let modelUrn:String = virtualDevice.getDeviceModel().getURN()
        guard let view:String = deviceViewData[modelUrn]?.view else {
            showError(controller: self, alertTitle: "Device Selection Alert",
                      message: "Cannot determing monitor view from \(modelUrn)",
                      actionTitle: "OK",
                      handler: nil)
            return
        }
        showMonitorView(view:view,
                        device:device, virtualDevice:virtualDevice)
    }
    
    func showMonitorView(view:String,
                         device:Device, virtualDevice:VirtualDevice) {
        DispatchQueue.main.async(execute: {
            let view = self.storyboard!.instantiateViewController(
                withIdentifier: view)
            
            // Pass information to the DeviceMonitorView
            (view as! DeviceMonitorView).enterpriseClientSample = self.enterpriseClientSample
            (view as! DeviceMonitorView).device = device
            (view as! DeviceMonitorView).virtualDevice = virtualDevice
            
            self.navigationController?.pushViewController(view,
                                                          animated: false)
        })
    }
    
    @objc func longShowId(sender: UILongPressGestureRecognizer) {
        guard sender.state == UIGestureRecognizer.State.began else {
            return
        }
        let touchPoint = sender.location(in: self.tableView)
        guard let indexPath = self.tableView.indexPathForRow(at: touchPoint),
                indexPath.row != 0,
                let cell:DeviceSelectionCell =
                    self.tableView.cellForRow(at: indexPath) as? DeviceSelectionCell,
                let virtualDevice:VirtualDevice = cell.virtualDevice else {
            return
        }
        let deviceId:String = virtualDevice.getEndpointId()
        showError(controller: self, alertTitle: "Device ID",
                      message: deviceId , actionTitle: "OK",
                      handler: nil)
    }
}


