/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// The DeviceMonitorView is common to all device monitor views. It
// encapsulates common properties for instances of EnterpriseClientSample,
// the selected Device and VirtualDevice

class EddystoneDeviceMonitorView: DeviceMonitorView {

    @IBOutlet var eddyStoneDeviceMonitorOutlet: UIView!
    @IBOutlet weak var deviceIdLabelOutlet: UILabel!
    @IBOutlet weak var manufacturerLabelOutlet: UILabel!
    @IBOutlet weak var modelNumberLabelOutlet: UILabel!
    @IBOutlet weak var serialNumberLabelOutlet: UILabel!
    @IBOutlet weak var deviceNamespaceLabelOutlet: UILabel!
    @IBOutlet weak var deviceInstanceLabelOutlet: UILabel!
    @IBOutlet weak var measuredSignalLabelOutlet: UILabel!
    @IBOutlet weak var receivedSignalLabelOutlet: UILabel!
    @IBOutlet weak var deviceTempLabelOutlet: UILabel!
    @IBOutlet weak var deviceVoltageLabelOutlet: UILabel!
    
    let rssiValueFormat:String = "Received: %.1f"
    let ora_txPowerValueFormat:String = "Measured: %d"
    let tempValueForamt:String = "Temperature: %.3f \u{00B0}C"
    let voltageValueFormat:String = "Battery Voltage; %.3f V"
    let namespaceValueFormat:String = "Namespace: %@"
    let instanceValueFormat:String = "Instance: %@"
    
    let majorValueFormat:String = "Major: %@"
    let minorValueFormat:String = "Minor: %@"

    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Monitor the selected device for Eddystone Beacon device model and
        // update the UI view accordingly.
        guard let virtualDevice:VirtualDevice = virtualDevice else {
            print("EddystoneDeviceMonitorView.viewDidLoad(): virtualDevice is nil")
            // TODO: Can this be done here ?
            _ = navigationController?.popViewController(animated: true)
            return
        }
            
        // Eddystone Beacon values changes are managed.
        virtualDevice.setOnChange(callback: { (event: ChangeEvent) in
            DispatchQueue.main.async(execute: {
                var namedValue = event.getNamedValue()
                repeat  {
                    switch namedValue.getName() {
                    case "ora_rssi":
                        let rssi:Double = (namedValue.getValue() as! NSNumber).doubleValue
                        self.receivedSignalLabelOutlet.text = String(format:self.rssiValueFormat, rssi)
                        break
                    case "ora_txPower":
                        let mssi:Int = (namedValue.getValue() as! NSNumber).intValue
                        self.measuredSignalLabelOutlet.text = String(format:self.ora_txPowerValueFormat, mssi)
                        break
                    case "temperature":
                        let temperature:Double = (namedValue.getValue() as! NSNumber).doubleValue
                        let celsiusTemp:Double = temperature / 256.0
                        self.deviceTempLabelOutlet.text = String(format:self.tempValueForamt, celsiusTemp)
                        break
                    case "batteryVoltage":
                        let voltage:Double = (namedValue.getValue() as! NSNumber).doubleValue
                        let voltValue:Double = voltage / 1000.0
                        self.deviceVoltageLabelOutlet.text = String(format:self.voltageValueFormat, voltValue)
                        break
                    default:
                        break
                    }
                    guard let next:NamedValue = namedValue.next() else {
                        break
                    }
                    namedValue = next
                }
                while true
            })
        })
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    override func viewWillAppear (_ animated: Bool) {
        
        deviceIdLabelOutlet.text = self.virtualDevice?.getEndpointId()
        
        manufacturerLabelOutlet.text = "Manufacturer: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MANUFACTURER.rawValue)) ?? "")
        modelNumberLabelOutlet.text = "Model Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MODEL_NUMBER.rawValue)) ?? "")
        serialNumberLabelOutlet.text = "Serial Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? "")
        
        let metadata:[String:String]? = device?.getMetadata()
        let uuid:String? = metadata?["UUID"]
        let tokens:[String]? = uuid?.components(separatedBy:":")
        self.deviceNamespaceLabelOutlet.text = String(format:namespaceValueFormat, tokens?[0] ?? "")
        self.deviceInstanceLabelOutlet.text = String(format:instanceValueFormat,
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? ""))
        
        do {
            let power:Int =
                (try virtualDevice?.get(attributeName: "ora_txPower") as? NSNumber)?.intValue ?? 0
            self.measuredSignalLabelOutlet.text = String(format:ora_txPowerValueFormat, power)
            let rssi:Double =
                (try virtualDevice?.get(attributeName: "ora_rssi") as? NSNumber)?.doubleValue ?? 0.0
            self.receivedSignalLabelOutlet.text = String(format:rssiValueFormat, rssi)
            let temp:Double =
                (try virtualDevice?.get(attributeName: "temperature") as? NSNumber)?.doubleValue ?? 0.0
            self.deviceTempLabelOutlet.text =
                String(format: tempValueForamt, temp/256.0)
            let voltage:Double =
                (try virtualDevice?.get(attributeName: "batteryVoltage") as? NSNumber)?.doubleValue ?? 0.0
            self.deviceVoltageLabelOutlet.text = String(format: voltageValueFormat, voltage/1000.0)
        } catch {
            print(error)
        }
    }
}
