/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import EnterpriseLib

/* The device models supported by this Sample */
let HUMIDITY_SENSOR_DEVICE_MODEL_URN = "urn:com:oracle:iot:device:humidity_sensor"
let TEMPERATURE_SENSOR_DEVICE_MODEL_URN = "urn:com:oracle:iot:device:temperature_sensor"
let IBEACON_DEVICE_MODEL_URN = "urn:com:oracle:iot:device:location:ibeacon"
let EDDYSTONEBEACONTLMUID_DEVICE_MODEL_URN = "urn:com:oracle:iot:device:location:eddystone-tlm-uid"
let MOTIONACTIVATEDCAMERA_DEVICE_MODEL_URN = "urn:com:oracle:iot:device:motion_activated_camera"

let supportedDeviceModels:[String] = [
    HUMIDITY_SENSOR_DEVICE_MODEL_URN,
    TEMPERATURE_SENSOR_DEVICE_MODEL_URN,
    IBEACON_DEVICE_MODEL_URN,
    EDDYSTONEBEACONTLMUID_DEVICE_MODEL_URN,
    MOTIONACTIVATEDCAMERA_DEVICE_MODEL_URN
]


fileprivate let autherr = NSLocalizedString(
    "User Authentication failed. Session cookie has expired.",
    comment: "")
func isUserAuthenticationError(error:String) -> Bool {
    return error.range(of: autherr) != nil
}

// This class creates an EnterpriseClient instance and
// obtains all its DeviceModels.
class EnterpriseClientSample {

    // The EnterpriseClient instance
    var enterpriseClient:EnterpriseClient? = nil
    // A queue to safely store DeviceModel data
    let deviceModelQueue:DispatchQueue =
        DispatchQueue(label:"EnterpriseSample deviceModel queue",
                      qos:.utility)
    var deviceModels:[String:DeviceModel] = [String:DeviceModel]()
    
    func getDeviceModels() -> [String:DeviceModel]? {
        return deviceModelQueue.sync(execute: {
            return self.deviceModels
        })
    }
    // The supported DeviceModel urn's
    var deviceModelUrns:[String] = [String]()
    
    // init() returns with an EnterpriseClientSample that may not
    // be complete and usable. Address this in the caller's callback method.
    // When that callback is invoked the EnterpriseClient instance has been
    // initialized, unless there is an error, and then its state is undefined.
    init(callback: @escaping (String?)->(), applicationName:String?=nil) {
        func _callback(enterpriseClient:EnterpriseClient?, error:ClientError?) {
            if let error = error {
                // An error has encountered during enterprise client creation.
                callback("\(error)")
                return
            }
            guard let enterpriseClient = enterpriseClient else {
                callback("EnterpriseClient.init(): newClient() " +
                    "return nil error and nil EnterpriseClient instance.")
                return
            }
            self.enterpriseClient = enterpriseClient
            // If the name is empty there has been an error
            if self.enterpriseClient?.getApplication().getName().isEmpty ?? true {
                callback("EnterpriseClient.init(): " +
                    "Enterprise Client object has not been created.")
                return
            }

            self.getDeviceModels(callback:{error in
                callback(error)
            })
            
        }
        // If application name is nil or empty use endpoint authentication and
        // pass the Provisioning information.
        if let applicationName = applicationName, !applicationName.isEmpty {
            // Create the enterprise client using application name.
            EnterpriseClient.newClient(appName: applicationName,
                                       callback: _callback)
        } else {
            // Create the enterprise client using file provisioning information.
            EnterpriseClient.newClient(path:Provisioning.getProvisioningFilePath(),
                                       password:Provisioning.getProvisioningPassword(),
                                       callback:_callback)
        }
    }

    
    // Process the device model recursively.
    // Accumulate all the device model urns first. This prevents
    // making network calls while processing the DeviceModelIteraor's
    // pages. The network call will return immediately and the next
    // page of device model urns make be processed before the first
    // page completes, complicating the process.
    
    var deviceModelCount:Int = 0
    func getDeviceModels(callback: @escaping (String?)->()) {
        // Process the device model recursively.
        // Accumulate all the device model urns first. This prevents
        // making network calls while processing the DeviceModelIteraor's
        // pages. The network call will return immediately and the next
        // page of device model urns make be processed before the first
        // page completes, complicating the process.
        func recursivePaging(pageable:Pageable<String>,
                              error: ClientError?) {
            let calledBackQueue:DispatchQueue =
                DispatchQueue(label:"EnterpriseSample calledBack queue",
                              qos:.utility)
            // A flag to track if the application callback is called,
            var calledBack:Bool = false
            // Since the "recursive" processing of device models happens
            // effectively in parallel, due to immediate return from asynchronous
            // network calls, ensure that the callback only happens once.
            func _callback(error:String?) {
                calledBackQueue.sync() {
                    if calledBack {
                        return
                    }
                    callback(error)
                    calledBack = true
                }
            }
            
            // This funtion accumulates the device model urns first, recursively.
            // Once all the urns are obtained, a DeviceModel instance is obtained
            // for each urn.
            // The callback can be called directly since these calls to _recursivePaging
            // are called one after the other, and each exits immediately except the last.
            func _recursivePaging(pageable:Pageable<String>?, error:ClientError?) {
                // If there is an error, stop processing pages.
                if let error = error {
                    callback("\(error)")
                    return
                }
                // In case both pageable and error nil, stop processing pages.
                guard let pageable = pageable else {
                    callback("EnterpriseClientSample.getDeviceModels(): " +
                        "_recursivePaging(): both error and pageable are nil")
                    return
                }
                // We know there is a valid page of results
                // Store the current page and get the next if there are more
                // or else stop processing pages
                deviceModelUrns.append(contentsOf:pageable.elements())
                if pageable.hasMore() {
                    do {
                        try pageable.next(callback: _recursivePaging)
                    } catch {
                        callback("\(error)")
                    }
                } else {
                    // Only the last call to _recursivePaging makes it to this point.

                    // we have all the device model urns, obtain the
                    // DeviceModel instances for each urn.
                    for urn:String in deviceModelUrns {
                        // Display only the device models supported by this application
                        guard supportedDeviceModels.contains(urn) else {
                            continue
                        }
                        do {
                            // Get DeviceModel instance for urn.
                            try self.enterpriseClient?.getDeviceModel(deviceModelUrn: urn,
                                                                      callback: { deviceModel, error in
                                // This section of code may be running in parallel with
                                // other recursive calls to recursivePaging, after all the
                                // DeviceModel urns have been obtained.
                                // Call the application callback if there is an error.
                                if let error = error {
                                    _callback(error:"\(error)")
                                    return
                                }
                                // Store the DeviceModel instance and its name.
                                // If the number of stored DeviceModel instances is equal
                                // to the total number of supported DeviceModels, call the
                                // application callback.
                                if let deviceModel = deviceModel {
                                    self.deviceModelQueue.sync() {
                                        self.deviceModels[deviceModel.getName()] = deviceModel
                                    }
                                }
                            })
                        } catch {
                            _callback(error:"\(error)")
                            return
                        }
                    }
                }
            }
            // Begin the processing of DeviceModelIterator pages.
            _recursivePaging(pageable:pageable, error:error)
        }
        
        // Get device model list associated to the application.
        self.enterpriseClient?.getDeviceModels(callback: { (pageable, error) in
            if let error = error {
                callback("\(error)")
                return
            }
            guard let pageable = pageable else {
                callback("EnterpriseClient.init(): getDeviceModels() " +
                    "returned nil error and nil DeviceModels pageable.")
                return
            }
            // No device models
            // Without device models the sample cannot do much.
            if pageable.size() <= 0 {
                callback("EnterpriseClientSample.getDeviceModels(): " +
                    "getDeviceModels() returned no models.")
                return
            }
            // The total count of device models
            self.deviceModelCount = pageable.size()
            recursivePaging(pageable:pageable, error:error)
        })
    }

    // Return the devices for a given device model
    func getDevices(deviceModel:DeviceModel,
                    callback: @escaping ([(Device,VirtualDevice)]?,String?)->()) {

        var _devices:[Device] = [Device]()
        // This function accumulates the devices first, recursively.
        // Once all the devices are obtained, a VirtualDevice instance is created
        // for each device.
        func recursivePaging(pageable:Pageable<Device>?, error:ClientError?) {
            // If there is an error, stop processing pages.
            if let error = error {
                callback(nil, "\(error)")
                return
            }
            // In case both pageable and error nil, stop processing pages.
            guard let pageable = pageable else {
                callback(nil, "EnterpriseSample.getDevices(): recursivePaging() " +
                    "returned error and pageable equal nil")
                return
            }
            // We know there is a valid page of results
            // Store the current page and get the next if there are more
            // or else stop processing pages
            _devices.append(contentsOf:pageable.elements())
            if pageable.hasMore() {
                do {
                    try pageable.next(callback: recursivePaging)
                } catch {
                    callback(nil, "EnterpriseSample.getDevices(): " +
                        "pageable.next() threw exception \(error)")
                }
            } else {
                // we have all the devices create a VirtualDevice for each device.
                // DeviceModel instances for each urn.
                var devices:[(Device,VirtualDevice)] = [(Device,VirtualDevice)]()
                for device:Device in _devices {
                    let deviceId:String = device.getId()
                    guard let virtualDevice:VirtualDevice =
                            self.enterpriseClient?.createVirtualDevice(deviceId: deviceId,
                                                                       deviceModel: deviceModel) else {
                        callback(nil, "EnterpriseSample.getDevices(): " +
                            "Cannot create VirtualDevice for device model " +
                            "\(deviceModel.getName()) with device id \(deviceId)")
                        return
                    }
                    devices.append((device,virtualDevice))
                }
                callback(devices, nil)
            }
        }

        // Get active device list for the selected URN.
        self.enterpriseClient?.getActiveDevices(deviceModel: deviceModel.getURN(),
                                                callback: { (pageable, error) in
            if let error = error {
                callback(nil, "\(error)")
                return
            }
            guard let pageable = pageable else {
                callback(nil, "EnterpriseClient.getDevices(): getActiveDevices() " +
                    "returned nil error and nil Device pageable.")
                return
            }
            recursivePaging(pageable: pageable, error: error)
        })
    }

    func close(callback: @escaping (String?)->()) {
        do {
            try self.enterpriseClient?.close()
            callback(nil)
        } catch {
            callback("\(error)")
        }
    }
}
