/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// The DeviceMonitorView is common to all device monitor views. It
// encapsulates common properties for instances of EnterpriseClientSample,
// the selected Device and VirtualDevice

class HumidityDeviceMonitorView: DeviceMonitorView, UITextFieldDelegate {
    
    @IBOutlet var humidityDeviceMonitorOutlet: UIView!
    @IBOutlet var deviceIdLabelOutlet: UILabel!
    @IBOutlet weak var manufacturerLabelOutlet: UILabel!
    @IBOutlet weak var modelNumberLabelOutlet: UILabel!
    @IBOutlet weak var serialNumberLabelOutlet: UILabel!
    @IBOutlet weak var humidityThresholdLabelOutlet: UILabel!
    @IBOutlet weak var humidityValueLabelOutlet: UILabel!
    @IBOutlet weak var humidityAlertLabelOutlet: UILabel!
    @IBOutlet weak var newThresoldTextFieldOutlet: UITextField!
    @IBOutlet var applyButtonOutlet: UIButton!
    
    static let DEFAULTTHRESHOLD:Int = 80
    let humidityValueFormat:String = "Humidity: %d %"
    let maxThresholdValueFormat:String = "Threshold: %d %"
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.newThresoldTextFieldOutlet.delegate = self
        self.title = NAVBARTITLE
        
        // Monitor the selected device for humidity device model and update the 
        // UI view accordingly.
        guard let virtualDevice = virtualDevice else {
            print("HumidityDeviceMonitorView.viewDidLoad(): virtualDevice is nil")
            _ = navigationController?.popViewController(animated: false)
            return
        }
        // Humidity values changes are managed.
        virtualDevice.setOnChange(callback: { (event: ChangeEvent) in
            DispatchQueue.main.async(execute: {
                var namedValue:NamedValue = event.getNamedValue()
                repeat  {
                    let attr:String = namedValue.getName()
                    switch attr {
                    case "humidity":
                        let humidity:Int = (namedValue.getValue() as! NSNumber).intValue
                        self.humidityValueLabelOutlet.text =
                            String(format:self.humidityValueFormat, humidity)
                        break
                    case "maxThreshold":
                        let maxThreshold:Int = (namedValue.getValue() as! NSNumber).intValue
                        self.humidityThresholdLabelOutlet.text =
                            String(format:self.maxThresholdValueFormat, maxThreshold)
                        break
                    default:
                        break
                    }
                    guard let next:NamedValue = namedValue.next() else {
                        break
                    }
                    namedValue = next
                }
                while true
            })
        })
            
        // Humidity alerts are managed.
        virtualDevice.setOnAlert(callback: { (event: AlertEvent) in
            DispatchQueue.main.async(execute: {
                let longUrn = event.getURN()
                if longUrn.hasSuffix(":too_humid") {
                    DispatchQueue.main.async(execute: {
                        self.humidityAlertLabelOutlet.text = "Too Humid"
                        self.humidityAlertLabelOutlet.isHidden = false
                        // Loop for sample measurements
                        // Timers must be started on the main thread.
                        // This code is executed from the Alamofire response
                        // queue. This queue has been changed to be NOT the main
                        // queue.
                        let _ = Timer.scheduledTimer(timeInterval: 3,
                            target: self, selector: #selector(self.hideAlert),
                            userInfo: nil, repeats: false)
                    })
                }
            })
        })
            
        // Humidity max threshold update errors are managed.
        virtualDevice.setOnError(callback: { (event: ErrorEvent) in
            DispatchQueue.main.async(execute: {
                let errMessage = event.getMessage()
                print("Error message: " + errMessage)
                if isUserAuthenticationError(error:errMessage) {
                    self.displayServerLoginView()
                }
                
                let namedValue = event.getNamedValue()
                if namedValue.getName() == "maxThreshold" {
                    let maxThreshold:Int =
                        (namedValue.getValue() as! NSNumber).intValue
                    self.humidityThresholdLabelOutlet.text =
                        String(format:self.maxThresholdValueFormat, maxThreshold)
                }
            })
        })
    }
    
    override func viewWillAppear (_ animated: Bool) {
        hideAlert()
        deviceIdLabelOutlet.text = self.virtualDevice?.getEndpointId()
        
        manufacturerLabelOutlet.text = "Manufacturer: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MANUFACTURER.rawValue)) ?? "")
        modelNumberLabelOutlet.text = "Model Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MODEL_NUMBER.rawValue)) ?? "")
        serialNumberLabelOutlet.text = "Serial Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? "")

        do {
            let humidity:Int =
                (try virtualDevice?.get(attributeName: "humidity") as? NSNumber)?.intValue ?? 0
            humidityValueLabelOutlet.text =
                String(format:humidityValueFormat, humidity)
            let maxThreshold:Int =
                (try virtualDevice?.get(attributeName: "maxThreshold") as? NSNumber)?.intValue ??
                    HumidityDeviceMonitorView.DEFAULTTHRESHOLD
            humidityThresholdLabelOutlet.text =
                String(format:maxThresholdValueFormat, maxThreshold)
        } catch {
            print(error)
        }
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // actions

    // State variables for maxThreshold text field editing
    var newMaxThresholdString:String = ""
    var newMaxThresholdInt:Int = DEFAULTTHRESHOLD
    func textFieldDidBeginEditing(_ textField: UITextField) {
        newMaxThresholdString = ""
        newMaxThresholdInt = HumidityDeviceMonitorView.DEFAULTTHRESHOLD
        applyButtonOutlet.isEnabled = true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField,
                                reason: UITextField.DidEndEditingReason) {
        guard let _newMaxThresholdString:String = textField.text,
                !_newMaxThresholdString.isEmpty else {
            return
        }
        // Validate a valid integer
        guard let _newMaxThresholdInt:Int = Int(_newMaxThresholdString) else {
            showError(controller: self, alertTitle: "Threshold Setting Alert",
                      message: "Not an integer \(_newMaxThresholdString)",
                      actionTitle: "OK",
                      handler: nil)
            return
        }
        newMaxThresholdString = _newMaxThresholdString
        newMaxThresholdInt = _newMaxThresholdInt
    }
    
    func textFieldShouldReturn(_ userText: UITextField) -> Bool {
        userText.resignFirstResponder()
        userText.endEditing(true)
        return false;
    }
    
    // Action to apply changes on humidity threshold.
    @IBAction func applyButtonOutletAction(_ sender: UIButton) {
        newThresoldTextFieldOutlet.resignFirstResponder()
        newThresoldTextFieldOutlet.endEditing(true)
        if newMaxThresholdString.isEmpty {
            return
        }
        
        do {
            let _ = try self.virtualDevice?.set(attributeName: "maxThreshold",
                                    attributeValue: newMaxThresholdInt as AnyObject)
        } catch ClientError.argument(let error) {
            print(error)
            showError(controller: self, alertTitle: "Threshold Setting Alert",
                      message: error, actionTitle: "OK",
                      handler: nil)
            return
        } catch  {
            print(error)
            showError(controller: self, alertTitle: "Humidity Alert",
                      message: "\(error)", actionTitle: "OK",
                      handler: nil)
            return
        }
        
        // update the on screen values
        self.newThresoldTextFieldOutlet.text = ""
        self.humidityThresholdLabelOutlet.text =
            String(format:maxThresholdValueFormat, newMaxThresholdInt)
        applyButtonOutlet.isEnabled = false
    }
    
    // helpers
    
    @objc func hideAlert() {
        self.humidityAlertLabelOutlet.isHidden = true
    }
    
    // Display the ServerLoginView on UserAuthentication errors
    func displayServerLoginView() {
        let serverLoginView = self.storyboard?.instantiateViewController(
            withIdentifier: "ServerLoginView") as! ServerLoginView
        self.navigationController?.pushViewController(serverLoginView,
                                                      animated: false)
    }
    
    func displayDeviceSelectionView() {
        let deviceSelectionView = self.storyboard?.instantiateViewController(
            withIdentifier: "DeviceSelectionView") as! DeviceSelectionView
        self.navigationController?.pushViewController(deviceSelectionView,
                                                      animated: false)
    }
}
