/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// The DeviceMonitorView is common to all device monitor views. It
// encapsulates common properties for instances of EnterpriseClientSample,
// the selected Device and VirtualDevice

class MotionActivatedCameraView: DeviceMonitorView, UITextViewDelegate,
                                 UITextFieldDelegate {
    
    private class MACVSyncCallback : SyncCallback {
        var block:((SyncEvent)->())!
        override init() {
            super.init()
        }
        convenience init(block: @escaping (SyncEvent)->()) {
            self.init()
            self.block = block
        }
        override public func onSync(event:SyncEvent) {
            block(event)
        }
    }
    private var onChangeSyncCallback:MACVSyncCallback?
    private var dataSyncCallback:MACVSyncCallback?
    private var alertSyncCallback:MACVSyncCallback?
    private var errorSyncCallback:MACVSyncCallback?
    
    @IBOutlet var motionActivatedViewOutlet: UIView!
    @IBOutlet weak var scheduleRecordingTextFieldOutlet: UITextField!
    @IBOutlet weak var scheduleRecordingLabelOutlet: UILabel!

    @IBOutlet var deviceIdLabelOutlet: UILabel!
    @IBOutlet weak var manufacturerLabelOutlet: UILabel!
    @IBOutlet weak var modelNumberLabelOutlet: UILabel!
    @IBOutlet weak var serialNumberLabelOutlet: UILabel!
    @IBOutlet weak var recordTimeLabelOutlet: UILabel!
    @IBOutlet weak var recordTimeTextFieldOutlet: UITextField!

    @IBOutlet weak var downloadDirectory: UILabel!
    @IBOutlet weak var imageMessageOutlet: UITextView!
    @IBOutlet weak var recordButtonOutlet: UIButton!
   
    override func viewDidLoad() {
        super.viewDidLoad()
        
        imageMessageOutlet.delegate = self
        recordTimeTextFieldOutlet.delegate = self
        
        // Monitor the selected device for uploads and update the
        // UI view accordingly.
        guard let _ = virtualDevice else {
            print("MotionActivatedCameraView.viewDidLoad(): virtualDevice is nil")
            return
        }
        
        onChangeSyncCallback =
            MACVSyncCallback(block: { (event) in
                print("#### In onChange SyncCallback ####")
                self.syncEventMsg(cb:"onChange", event:event)
            })
        
        
        dataSyncCallback =
            MACVSyncCallback(block: { (event) in
                print("#### In Data SyncCallback ####")
                self.syncEventMsg(cb:"onData", event:event)
            })
        
        alertSyncCallback =
            MACVSyncCallback(block: { (event) in
                print("#### In Alert SyncCallback ####")
                self.syncEventMsg(cb:"onAlert", event:event)
            })
        
        
        errorSyncCallback =
            MACVSyncCallback(block: { (event) in
                print("#### In Error SyncCallback ####")
                self.syncEventMsg(cb:"onError", event:event)
            })

        registerCallbacks()
    }
    
    override func viewWillAppear (_ animated: Bool) {
        
        deviceIdLabelOutlet.text = self.virtualDevice?.getEndpointId()
        
        manufacturerLabelOutlet.text = "Manufacturer: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MANUFACTURER.rawValue)) ?? "")
        modelNumberLabelOutlet.text = "Model Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MODEL_NUMBER.rawValue)) ?? "")
        serialNumberLabelOutlet.text = "Serial Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? "")
    
        let id = getImageDir()
        let dd = String(id[getDocumentPath(file: "").endIndex..<id.endIndex])
        downloadDirectory.text = dd
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // actions
    
    @objc var canRecord:Bool = false
    @objc var canSchedule:Bool = false
    
    // Action to apply changes on humidity threshold.
    @IBAction func recordButtonAction(_ sender: UIButton) {
        // If we're here the record button has been enabled
        // and there should be data in one of the text fields.
        defer {
            canRecord = false
            recordTimeTextFieldOutlet.text = ""
            scheduleRecordingTextFieldOutlet.text = ""
            recordButtonOutlet.isEnabled = false
        }
        
        // Get the recording duration if set to greater than 0
        var recordTime:Int? = 0
        if let recordTimeString = recordTimeTextFieldOutlet.text {
            recordTime = Int(recordTimeString)
            if recordTime != nil && recordTime! > 0 {
                canRecord = true
            }
        }
        
        var scheduleRecording:Int? = 0
        if let scheduleRecordingString = scheduleRecordingTextFieldOutlet.text {
            scheduleRecording = Int(scheduleRecordingString)
            if scheduleRecording != nil && scheduleRecording! > 0 {
                canSchedule = true
            }
        }

        // If scheduleRecording has been set, call the scheduleRecording action and pass
        // the derived start time "now + scheduleRecording" and duration. If canRecord
        // is false pass 0 for duration
        do {
            defer {
                canRecord = false
                canSchedule = false
            }
            if canSchedule {
                let startTime:Int =
                    (Date(timeIntervalSinceNow: Double(scheduleRecording!))
                        .timeIntervalSince1970 as NSNumber).intValue
                
                if let action:Action = try self.virtualDevice?.createAction(actionName: "scheduleRecording") {
                    try action.set(argumentName: "duration", value: recordTime!)
                              .set(argumentName: "startTime", value: startTime)
                              .call()
                }
            } else if canRecord {
                let _ = try self.virtualDevice?.call(actionName: "record",
                                                     argument: recordTime!)
                
            }

        } catch {
            print(error)
            showError(controller: self, alertTitle: "Record Action Alert",
                message: "\(error)", actionTitle: "OK",
                handler: { (action:UIAlertAction?) -> () in
            })
        
        }
    }
    // If neither set just fall through, state will be reset.
    
    func textFieldShouldReturn(_ userText: UITextField) -> Bool {
        userText.resignFirstResponder()
        userText.endEditing(true)
        return false;
    }
    
    // TODO: Should just "grab" the "right" value here and not when the
    // TODO: record button action is triggered.
    func textFieldDidEndEditing(_ textField: UITextField) {
        // This probably will never happen ?
        let seconds:String? = textField.text
        if textField == recordTimeTextFieldOutlet {
            if seconds != nil && seconds! != "" {
                canRecord = true
                recordButtonOutlet.isEnabled = true
            } else {
                canRecord = false
            }
        } else if textField == scheduleRecordingTextFieldOutlet {
            if seconds != nil && seconds! != "" {
                canRecord = true
                recordButtonOutlet.isEnabled = true
            } else {
                canSchedule = false
            }
        }
        if !canRecord && !canSchedule {
            recordButtonOutlet.isEnabled = false
        }
    }
    
    func textFieldDidBeginEditing(_ textField: UITextField) {
        if textField == recordTimeTextFieldOutlet {
            canRecord = false
        } else if textField == scheduleRecordingTextFieldOutlet {
            canSchedule = false
        }
        recordButtonOutlet.isEnabled = false
    }

    private func update(text:String) {
        DispatchQueue.main.async(execute: {
            let outtext = self.imageMessageOutlet.text + "\n" + text
            self.imageMessageOutlet.text = outtext
            // Force text field to (invisibly) scroll until latest line is at bottom
            let bottom: NSRange = NSMakeRange(self.imageMessageOutlet.text.count - 10, 1)
            self.imageMessageOutlet.scrollRangeToVisible((bottom))
        })
    }
    
    @objc func registerCallbacks() {
        
        virtualDevice?.setOnChange(callback: { (event: ChangeEvent) in
            let msg = "onChange: "
            self.processNamedValue(prefix: msg, namedValue: event.getNamedValue(),
                                   soCB: self.onChangeSyncCallback!)
        })
        
        virtualDevice?.setOnAlert(callback: { (event: AlertEvent) in
            
            let longUrn = event.getURN()
            let etime = event.getEventTime()
            let index = longUrn.range(of: ":", options: .backwards)?.lowerBound
            let shortUrn = String(longUrn[index!..<longUrn.endIndex])
            
            let msgFormat = "onAlert: %@ : %@ : "
            let msg:String = String(format: msgFormat,
                                    ISO8601DateFormatter().string(from: etime),
                                    shortUrn)
            
            self.processNamedValue(prefix: msg, namedValue: event.getNamedValue(),
                                   soCB: self.alertSyncCallback!)
        })
        
        virtualDevice?.setOnError(callback: { (event: ErrorEvent) in
            let errMessage = event.getMessage()
            print("Error message: " + errMessage)
            if isUserAuthenticationError(error: errMessage) {
                // Should post alert popup
                DispatchQueue.main.async(execute: {
                    let firstViewController = self.storyboard?.instantiateViewController(
                        withIdentifier: "ServerLoginView") as! ServerLoginView
                    self.navigationController?.pushViewController(firstViewController,
                                                                  animated: false)
                })
                return
            }
            let errorFormat = "onError: %@ : "
            let msg = String(format: errorFormat, errMessage)
            self.processNamedValue(prefix: msg, namedValue: event.getNamedValue(),
                                   soCB: self.errorSyncCallback!)
        })
        
        // this callback will  be invoked when data is received for the format
        // urn:com:oracle:iot:device:motion_activated_camera:recording
        virtualDevice?.setOnData(formatURN: MOTIONACTIVATEDCAMERA_DEVICE_MODEL_URN + ":recording",
                                 callback: { event in
            let etime = event.getEventTime()
            // Just use recording the model urn is always the same.
            let urn = ":recording" // "event.getURN()
            
            let format:String = "onData: %@ : %@ : "
            let msg:String = String(format: format,
                                    ISO8601DateFormatter().string(from: etime),
                                    urn)
            self.processNamedValue(prefix: msg, namedValue: event.getNamedValue(),
                                   soCB: self.dataSyncCallback!)
        })
    }
    
    private func processNamedValue(prefix:String, namedValue:NamedValue,
                           soCB: MACVSyncCallback) {
        var msg = prefix + "\n"
        var _namedValue:NamedValue? = namedValue
        var sep = ""
        let format = "%@   \"%@\"=%@"
        while _namedValue != nil {
            let name:String = _namedValue!.getName()
            let value:Any = _namedValue!.getValue() ?? "nil"
            if name == "imageTime" {
                msg += String(format: format, sep, name,
                    ISO8601DateFormatter().string(from: Date(
                        timeIntervalSince1970: TimeInterval(truncating: value as! NSNumber))))
            } else if value is Date {
                msg += String(format: format, sep, name,
                              ISO8601DateFormatter().string(from: value as! Date))
            } else if value is StorageObject {
                // Oracle Cloud Storage
                let so:StorageObject = value as! StorageObject
                self.getStorageObject(so: so, callback: soCB)
                msg += String(format: format, sep, name, so.getName())
            } else if value is ExternalObject {
                // No or unknown storage server
                let eo:ExternalObject = value as! ExternalObject
                msg += String(format: format, sep, name, eo.getURI())
            } else {
                msg += String(format: format, sep, name, "\(value)")
            }
            _namedValue = _namedValue!.next();
            sep = "\n"
        }
        self.update(text: msg)
    }

    @objc let syncEventFormat = "SyncEvent(%@)\n  %@ name: %@\n  syncStatus: %@\n  path: %@"
    
    private func syncEventMsg(cb:String, event:SyncEvent) {
        let so:StorageObject = event.getSource()
        // This is composed of the "getImageDir" and the output file name
        // Get just the path substring, last path components
        let op = so.getOutputPath()!
        let filepath = String(op[getImageDir().endIndex..<op.endIndex])
        let msg = String(format: syncEventFormat, cb,
                         event.getName()!, so.getName(), "\(so.getSyncStatus())", filepath)
        self.update(text: msg)
    }

    // TODO: Should probably be triggered by a download button.
    private func getStorageObject(so:StorageObject, callback: MACVSyncCallback) {
        // The client library gives control to the application to decide
        // whether and where to download content from the Storage Cloud Service.
        so.setOnSync(callback: callback);
        do {
            try so.setOutputPath(outputPath: getImageDir() + so.getName())
            // sync happens in the background and calls the syncCallback when done
            try so.sync()
        } catch  {
            print(error)
        }
    }
}
