/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

// The ServerLoginView is not currently used.
// It's purpose is to support UserAuthentication.
// However because of the prevalence of file provisioning and endpoint
// authentication, the code must be modified to support UserAuthentication
// in favor of or in conjunction with file provisioning.
//
// Therefore the Main storyboard's entry point is the DeviceModelSelectionScreen.
// The ApplicationSelectionView is used in conjunction with User Authentication,
// is also unused at this time.
// In order for this to screen even to appear, the "Main Interface" in the
// "General" tab must be set to "Main".

import UIKit
import WebKit
import EnterpriseLib

// A WebView is used to display the IDCS sign in screen. The
// Authentication cookies are stored in webView.configuration.websiteDataStore.httpCookieStore
// and used from there in all requests from HttpSecureConnectionUserAuth

// Open issues.
// - setting the server url into the TrustedAssetsManager that is created
//   during the instantiation of the EnterpriseClient
// - where is the default IDCS server url assigned.

class ServerLoginView: UIViewController, WKNavigationDelegate {
    
    @IBOutlet var serverLoginViewOutlet: UIView!
    @IBOutlet weak var oracleLogoImageViewOutlet: UIImageView!
    @IBOutlet weak var urlTextFieldOutlet: UITextField!
    @IBOutlet weak var loadUrlButtonOutlet: UIButton!
    @IBOutlet weak var webViewOutlet: WKWebView!
    @IBOutlet weak var stopButtonOutlet: UIButton!
    
    @objc var loadStatus: String = ""
    @objc var redirectOAM: Bool = false
    @objc var redirectURL: String?
    @objc let defaults = UserDefaults.standard
    
    var requestObject:URLRequest?
    
    override func viewDidLoad() {
        super.viewDidLoad()
        //let url = URL (string: serverHostName)
        //urlTextFieldOutlet.text = url!.description
        webViewOutlet.scrollView.isScrollEnabled = true
        webViewOutlet.navigationDelegate = self
        self.title = NAVBARTITLE
    }
    
    override func viewWillAppear (_ animated: Bool) {
        //self.navigationItem.setHidesBackButton(true, animated: false)
        super.viewWillAppear(animated)
        
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.addObserver(self, selector: #selector(self.terminate),
                                               name: UIApplication.willTerminateNotification, object: nil)
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        NotificationCenter.default.removeObserver(self)
        super.viewWillDisappear(animated)
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // NOTE:
    // This WKNavigationDelegate "works" as far as that if given an ids
    // server url, it will load the login page, but it has not been tested
    // past that point, i.e. if the credentials are good it is not clear if
    // the following logic for the response to the login page is correct
    // and the rest of the app will function function as expected with
    // respect to authentication etc.
    
    func webView(_ webView:WKWebView, didCommit navigation: WKNavigation!) {
        // Called when the web view begins to receive web content.
        print("didCommit navigation")
        
        if let redirectUrl:String = self.redirectURL {
            if let redirectURL = self.redirectURL,
                    self.redirectOAM == true {
                defaults.setValue(redirectURL, forKey: "requestURL")
                defaults.synchronize()
            }
            
            if redirectUrl != self.urlTextFieldOutlet.text {
                self.loadStatus = "Redirect"
            } else {
                if self.loadStatus.isEmpty == false {
                    self.loadStatus = "Authenticated"
                }
            }
        }
    }

    func webView(_ webView:WKWebView,
                 didStartProvisionalNavigation navigation: WKNavigation!) {
        // Called when web content begins to load in a web view.
        print("didStartProvisionalNavigation navigation")
        
        if self.redirectOAM == false {
            self.redirectOAM = true
        }
    }

    func webView(_ webView: WKWebView, didFail navigation: WKNavigation!, withError error: Error) {
        // Called when an error occurs during navigation.
        print("didFail \(error)")
    }
    func webView(_ webView: WKWebView,
                 didFailProvisionalNavigation navigation: WKNavigation!,
                 withError error: Error) {
        // Called when an error occurs while the web view is loading content.
        print("didFailProvisionalNavigation \(error)")
    }


    func webView(_ webView:WKWebView,
                 didReceiveServerRedirectForProvisionalNavigation navigation: WKNavigation!) {
        //Called when a web view receives a server redirect.
        print("didReceiveServerRedirectForProvisionalNavigation")
        print("RedirectURL? " + (webView.url?.absoluteString ?? "nil"))
        self.redirectURL = webView.url?.absoluteString
    }
    
    func webView(_ webView: WKWebView,
                 didReceive challenge: URLAuthenticationChallenge,
                 completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        // Called when the web view needs to respond to an authentication challenge.
        print("didReceive challenge")
        completionHandler(.performDefaultHandling, nil)
        //let cred:URLCredential = URLCredential(user: <some user name>,
        //                                       password: <some password>,
        //                                       persistence: .forSession)
        //completionHandler(.useCredential, cred)
    }
 
    func webView(_ webView: WKWebView,
                          didFinish navigation: WKNavigation!) {
        // Called when the navigation is complete.
        print("didFinish navigation")
        
        if self.loadStatus == "Authenticated" {
            // Get cookies
            webView.configuration.websiteDataStore.httpCookieStore.getAllCookies { cookies in
                for cookie in cookies {
                    print("\(cookie.name) is set to \(cookie.value)")
                }
            }
            // Go to application selection screen
            let applicationSelectionView = self.storyboard?.instantiateViewController(
                withIdentifier: "ApplicationSelectionView") as! ApplicationSelectionView
            self.navigationController?.pushViewController(applicationSelectionView,
                                                          animated: true)
        } else {
            if self.loadStatus.isEmpty == true {
                // Get cookies
                webView.configuration.websiteDataStore.httpCookieStore.getAllCookies { cookies in
                    for cookie in cookies {
                        print("\(cookie.name) is set to \(cookie.value)")
                        webView.configuration.websiteDataStore.httpCookieStore.delete(cookie,
                                                                                      completionHandler: nil)
                    }
                }

                let getURL = defaults.string(forKey: "requestURL")
                defaults.setValue(self.redirectURL, forKey: "requestURL")
                defaults.synchronize()

                let url = URL (string: getURL!)
                let requestObject = URLRequest(url: url!)
                webViewOutlet.load(requestObject)
            }
        }
    }
    
    func webViewWebContentProcessDidTerminate(_ webView: WKWebView) {
        // Called when the web view’s web content process is terminated.
        print("webViewWebContentProcessDidTerminate")
    }
    /*
    func webView(_ webView: WKWebView,
                 decidePolicyFor navigationAction: WKNavigationAction,
                 decisionHandler: @escaping (WKNavigationActionPolicy) -> Void) {
        // Decides whether to allow or cancel a navigation.
        print("decidePolicyFor navigationAction")
        decisionHandler(.allow)
    }
    
    func webView(_ webView: WKWebView,
                 decidePolicyFor navigationResponse: WKNavigationResponse,
                 decisionHandler: @escaping (WKNavigationResponsePolicy) -> Void) {
        // Decides whether to allow or cancel a navigation after its response is known.
        print("decidePolicyFor navigationResponse")
        decisionHandler(.allow)
    }
    */
    
    // actions
    // Load URL button action for user authentication 
    @IBAction func loadUrlButtonOutletAction(_ sender: UIButton) {
        if let urlstring:String = urlTextFieldOutlet.text,
                let url:URL = URL(string: urlstring) {
            let _requestObject:URLRequest = URLRequest(url: url)
            webViewOutlet.load(_requestObject)
            requestObject = _requestObject
        }
    }
    
    @IBAction func stopButtonOutletAction(_ sender: UIButton) {
        // When the app is suspended, because of the Info.plist property
        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
        // observer is called and the "terminate" method will be called.
        UIControl().sendAction(#selector(NSXPCConnection.suspend),
                               to: UIApplication.shared, for: nil)
    }
    
    @objc func terminate() {
        NotificationCenter.default.removeObserver(self)
        saveSettingsFile()
    }
    
}
