/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import EnterpriseLib

// The DeviceMonitorView is common to all device monitor views. It
// encapsulates common properties for instances of EnterpriseClientSample,
// the selected Device and VirtualDevice

class TemperatureDeviceMonitorView: DeviceMonitorView, UITextFieldDelegate {

    @IBOutlet var TemperatureDeviceMonitorViewOutlet: UIView!
    @IBOutlet weak var deviceIdLabelOutlet: UILabel!
    
    @IBOutlet weak var manufacturerLabelOutlet: UILabel!
    @IBOutlet weak var modelNumberLabelOutlet: UILabel!
    @IBOutlet weak var serialNumberLabelOutlet: UILabel!
    
    @IBOutlet weak var tempLabelOutlet: UILabel!
    @IBOutlet weak var alertLabelOutlet: UILabel!
    @IBOutlet weak var minTempLabelOutlet: UILabel!
    @IBOutlet weak var maxTempLabelOutlet: UILabel!
    @IBOutlet weak var thresholdValuesOutlet: UILabel!
    @IBOutlet weak var resetButtonOutlet: UIButton!
    @IBOutlet weak var newMinTextFieldOutlet: UITextField!
    @IBOutlet weak var newMaxTextFieldOutlet: UITextField!
    @IBOutlet weak var applyButtonOutlet: UIButton!
    
    @IBOutlet weak var lastOnLabelOutlet: UILabel!
    @IBOutlet weak var powerSwitchOutlet: UISwitch!
    
    let thresholdValuesFormat:String = "Threshold Values: Min %d Max %d %@"
    let temperatureValueFormat:String = "Temperature: %.2f %@"
    let minTempValueFormat = "Minimum: %.2f %@"
    let maxTempValueFormat = "Maximum: %.2f %@"
    let lastOnValueFormat = "Last ON: %@"
    
    // These are class level because the display of the values is a
    // single string, so if only one changes, we need the value of the
    // one that hasn't changed.
    var currentMaxThreshold:Int = 70
    var currentMinThreshold:Int = 0
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.title = NAVBARTITLE
        self.newMinTextFieldOutlet.delegate = self
        self.newMaxTextFieldOutlet.delegate = self

        // Monitor the selected device for temperature device model and update 
        // the UI view accordingly.
        guard let virtualDevice = virtualDevice else {
            print("TemperatureDeviceMonitorView.viewDidLoad(): virtualDevice is nil")
            _ = navigationController?.popViewController(animated: false)
            return
        }
        
        let units:String = getUnits()
        
        // TODO: What happens if the device is powered off there an event notification ?
        // Temperature values changes are managed.
        virtualDevice.setOnChange(callback: { (event: ChangeEvent) in
            var namedValue = event.getNamedValue()
            repeat {
                let attValue = namedValue.getValue()
                DispatchQueue.main.async() {
                    switch namedValue.getName() {
                        case "temp":
                            let temperature = (attValue as! NSNumber).doubleValue

                            self.tempLabelOutlet.text = String(format:self.temperatureValueFormat,
                                temperature, units)
                            break;
                        case "maxThreshold":
                            self.currentMaxThreshold = (attValue as! NSNumber).intValue
                            break
                        
                        case "minThreshold":
                            self.currentMinThreshold = (attValue as! NSNumber).intValue
                            break
                        
                        case "minTemp":
                            let minTemperature = (attValue as! NSNumber).doubleValue
                            self.minTempLabelOutlet.text =
                                String(format: self.minTempValueFormat, minTemperature, units)
                            break
                        case "maxTemp":
                            let maxTemperature = (attValue as! NSNumber).doubleValue
                            self.maxTempLabelOutlet.text =
                                String(format: self.maxTempValueFormat, maxTemperature, units)
                            break
                        case "startTime":
                            let startTime = (attValue as! NSNumber).doubleValue
                            let startDate = Date(timeIntervalSince1970: startTime)
                            let lastOn = formatDate(date:startDate)
                            self.lastOnLabelOutlet.text = String(format:self.lastOnValueFormat, lastOn)
                            break
                        default:
                            break
                    }
                } // end DispatchQueue.async
                guard let next:NamedValue = namedValue.next() else {
                    break
                }
                namedValue = next
            }
            while true
            
            DispatchQueue.main.async() {
                self.thresholdValuesOutlet.text = String(format:self.thresholdValuesFormat,
                    self.currentMinThreshold, self.currentMaxThreshold, units)
            }
            
        })
        
        // Temperature alerts are managed.
        virtualDevice.setOnAlert(callback: { (event: AlertEvent) in
            let longUrn = event.getURN()
            // Loop for sample measurements
            // Timers must be started on the main thread.
            // This code is executed from the Alamofire response
            // queue. This queue has been changed to be NOT the main
            // queue.
            DispatchQueue.main.async() {
                if longUrn.hasSuffix(":too_cold") {
                    self.alertLabelOutlet.text = "Too Cold"
                    self.alertLabelOutlet.backgroundColor = .blue
                    self.alertLabelOutlet.textColor = .white
                    self.alertLabelOutlet.isHidden = false
                }
                if longUrn.hasSuffix(":too_hot") {
                    self.alertLabelOutlet.text = "Too Hot"
                    self.alertLabelOutlet.backgroundColor = .red
                    self.alertLabelOutlet.textColor = .black
                    self.alertLabelOutlet.isHidden = false
                }
                let _ = Timer.scheduledTimer(timeInterval: 3, target: self,
                                         selector: #selector(self.hideAlert),
                                         userInfo: nil, repeats: false)
            } // end DispatchQueue.async
        })
        
        // Temperature min and max threshold update errors are managed.
        virtualDevice.setOnError(callback: { (event: ErrorEvent) in
            let errMessage = event.getMessage()
            print("Error message: " + errMessage)
            if isUserAuthenticationError(error:errMessage) {
                self.displayServerLoginView()
            }

            let namedValue = event.getNamedValue()
            DispatchQueue.main.async() {
                if namedValue.getName() == "maxThreshold" {
                    self.currentMaxThreshold = (namedValue.getValue() as! NSNumber).intValue
                }
                if namedValue.getName() == "minThreshold" {
                    self.currentMinThreshold = (namedValue.getValue() as! NSNumber).intValue
                }
                self.thresholdValuesOutlet.text = String(format:self.thresholdValuesFormat,
                                                         self.currentMinThreshold, self.currentMaxThreshold, units)
            } // end DispatchQueue.main.async
        })
    }

    override func viewWillAppear (_ animated: Bool) {
        // Get the current values whenver the view is shown
        hideAlert()
        deviceIdLabelOutlet.text = self.virtualDevice?.getEndpointId()
        
        manufacturerLabelOutlet.text = "Manufacturer: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MANUFACTURER.rawValue)) ?? "")
        modelNumberLabelOutlet.text = "Model Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.MODEL_NUMBER.rawValue)) ?? "")
        serialNumberLabelOutlet.text = "Serial Number: " +
            (device?.getValue(field: Device.Field.fromAlias(alias: Device.Field.Values.SERIAL_NUMBER.rawValue)) ?? "")

        do {
            let units:String = getUnits()
            
            let temperature:Double =
                (try virtualDevice?.get(attributeName: "temp") as? NSNumber)?.doubleValue ?? 0
            self.tempLabelOutlet.text = String(format:temperatureValueFormat, temperature, units)

            self.currentMinThreshold =
                (try virtualDevice?.get(attributeName: "minThreshold") as? NSNumber)?.intValue ?? 0
            self.currentMaxThreshold =
                (try virtualDevice?.get(attributeName: "maxThreshold") as? NSNumber)?.intValue ?? 70
            
            self.thresholdValuesOutlet.text = String(format:self.thresholdValuesFormat,
                self.currentMinThreshold, self.currentMaxThreshold, getUnits())
            
            let minTemp:Double =
                (try virtualDevice?.get(attributeName: "minTemp") as? NSNumber)?.doubleValue ?? 0.0
            self.minTempLabelOutlet.text = String(format: minTempValueFormat, minTemp, units)
            let maxTemp:Double =
                (try virtualDevice?.get(attributeName: "maxTemp") as? NSNumber)?.doubleValue ?? 70.0
            self.maxTempLabelOutlet.text = String(format: maxTempValueFormat, maxTemp, units)
            
            if let startTime = (try virtualDevice?.get(attributeName: "startTime") as? NSNumber)?.doubleValue {
                let startDate:Date = Date(timeIntervalSince1970: startTime)
                let lastOn = formatDate(date:startDate)
                self.lastOnLabelOutlet.text = String(format:lastOnValueFormat, lastOn)
            }

        } catch {
            print(error)
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        inTextField = false
    }
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    // actions
    
    // Action to apply power off/on for temperature device.
    @IBAction func powerSwitchAction(_ sender: UISwitch) {
        do {
            try self.virtualDevice?.call(actionName: "power",
                                         argument: sender.isOn as AnyObject)
        } catch {
            print(error)
            showError(controller: self,
                      alertTitle: "Power Action Error",
                      message: "\(error)",
                      actionTitle: "OK", handler: nil)
        }
    }

    // Action to reset temperature device.
    @IBAction func resetButtonAction(_ sender: AnyObject) {
        do {
            try self.virtualDevice?.call(actionName: "reset")
        } catch {
            print(error)
            showError(controller: self, alertTitle: "Reset Action Error",
                      message: "\(error)", actionTitle: "OK", handler: nil)
        }
    }
    
    // State variables for threshold text field editing
    var newMaxThresholdString:String = ""
    var newMaxThresholdInt:Int = -1
    var newMinThresholdString:String = ""
    var newMinThresholdInt:Int = -1
    
    func resetMaxTextField() {
        newMaxThresholdString = ""
        newMaxThresholdInt = -1
        newMaxTextFieldOutlet.text = ""
    }
    func resetMinTextField() {
        newMinThresholdString = ""
        newMinThresholdInt = -1
        newMinTextFieldOutlet.text = ""
        inTextField = false
    }
    
    var inTextField:Bool = false
    func textFieldDidBeginEditing(_ textField: UITextField) {
        if textField == newMinTextFieldOutlet {
            resetMinTextField()
        }
        if textField == newMaxTextFieldOutlet {
            resetMaxTextField()
        }
        applyButtonOutlet.isEnabled = true
    }
    
    
    func textFieldDidEndEditing(_ textField: UITextField,
                                reason: UITextField.DidEndEditingReason) {
        if textField == newMinTextFieldOutlet {
            if let _newMinThresholdString:String = textField.text,
                    !_newMinThresholdString.isEmpty{
                newMinThresholdString = _newMinThresholdString
            }
        }
        
        if textField == newMaxTextFieldOutlet {
            if let _newMaxThresholdString:String = textField.text,
                    !_newMaxThresholdString.isEmpty {
                newMaxThresholdString = _newMaxThresholdString
            }
        }
        if newMaxThresholdString.isEmpty && newMinThresholdString.isEmpty {
            applyButtonOutlet.isEnabled = false
        }
    }
    
    func textFieldShouldReturn(_ userText: UITextField) -> Bool {
        userText.resignFirstResponder()
        userText.endEditing(true)
        return false;
    }
    
    // Action to apply changes on temperature thresholds.
    @IBAction func applyButtonAction(_ sender: UIButton) {
        newMinTextFieldOutlet.resignFirstResponder()
        newMinTextFieldOutlet.endEditing(true)
        newMaxTextFieldOutlet.resignFirstResponder()
        newMaxTextFieldOutlet.endEditing(true)
        
        func setAttribute(attribute:String, value:String, intValue: inout Int) -> Bool {
            guard let intvalue:Int = Int(value) else {
                showError(controller: self,
                          alertTitle: "Threshold Setting Alert: \(attribute)",
                    message: "\(value) is not an integer", actionTitle: "OK",
                    handler: nil)
                return false
            }
            do {
                let _ = try self.virtualDevice?.set(attributeName: attribute,
                                                    attributeValue: intvalue as AnyObject)
                intValue = intvalue
                return true
            } catch {
                print(error)
                showError(controller: self,
                          alertTitle: "Threshold Setting Alert: \(attribute)",
                          message: "\(error)", actionTitle: "OK",
                          handler: nil)
                return false
            }
        }

        if !newMinThresholdString.isEmpty &&
                setAttribute(attribute: "minThreshold",
                             value: newMinThresholdString, intValue: &newMinThresholdInt) {
            self.currentMinThreshold = newMinThresholdInt
        }
        
        if !newMaxThresholdString.isEmpty &&
                setAttribute(attribute: "maxThreshold",
                             value: newMaxThresholdString, intValue: &newMaxThresholdInt) {
            self.currentMaxThreshold = newMaxThresholdInt
        }
        self.thresholdValuesOutlet.text = String(format:self.thresholdValuesFormat,
                                                 self.currentMinThreshold, self.currentMaxThreshold, getUnits())
        applyButtonOutlet.isEnabled = false
        resetMinTextField()
        resetMaxTextField()
    }
    
    // helpers
    
    // Display the ServerLoginView in case of user authentication errors.
    func displayServerLoginView() {
        DispatchQueue.main.async() {
            let serverLoginView =
                self.storyboard?.instantiateViewController(
                    withIdentifier: "ServerLoginView") as! ServerLoginView
            self.navigationController?.pushViewController(serverLoginView,
                                                          animated: false)
        }
    }
    
    @objc func hideAlert() {
        self.alertLabelOutlet.isHidden = true
    }
    
    func getUnits() -> String {
        var units:String = "\u{00B0}F"
        do {
            switch try virtualDevice?.get(attributeName: "unit") as? String ?? "" {
            case "Cel", "Celsius", "\u{00B0}C":
                units = "\u{00B0}C"
            default:
                units = "\u{00B0}F"
            }
        } catch {
            print(error)
        }
        return units
    }
}
