/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */
import Foundation
import DeviceLib
import os

public class Camera {
    var hardwareId:String!
    var description:String = "A motion activated camera"
    init(id:String) {
        self.hardwareId = id
    }
}

/**
 * All files uploaded to the storage cloud will be placed in a subdirectory 
 * named with the device endpoint id. The library configuration property
 * Config.DISABLE_STORAGE_OBJECT_PREFIX controls this behavior. By default it
 * is set to false.
 * If Config.DISABLE_STORAGE_OBJECT_PREFIX is set to true, all uploaded files
 * will be placed in the top level directory of the storage cloud.
 * To change Config.DISABLE_STORAGE_OBJECT_PREFIX the value must be changed
 * and the DeviceLib recompiled.
 *
 * Set the propery UNIQUE_FILENAME to true prevent overwriting files in the 
 * top level directory of the storage cloud when setting 
 * Config.DISABLE_STORAGE_OBJECT_PREFIX to true. This will add the device
 * endpoint id to the filename.
 */
let UNIQUE_FILENAME:Bool = false

let MOTION_ACTIVATED_CAMERA_MODEL_URN =
    "urn:com:oracle:iot:device:motion_activated_camera"
let IMAGE_ATTRIBUTE = "image"
let IMAGE_TIME_ATTRIBUTE = "imageTime"

let ORACLE_IOT_PNG = "oracle_iot.png"

/* The Logger
 *
 * An example of using the log command line tool to stream the
 * running application log output.
 *
 * $ sudo log stream --level default --predicate \
 *  'subsystem == "com.oracle.iot.sample" && category == "MotionActivatedCameraSample"'
 */
fileprivate let logger:OSLog = OSLog(subsystem: "com.oracle.iot.sample",
                                     category: "DirectlyConnectedDeviceSample")

@objc
public class MotionActivatedCamera : NSObject {
    
    class MACVSyncCallback : SyncCallback {
        var block:((SyncEvent)->())!
        override init() {
            super.init()
        }
        convenience init(block: @escaping (SyncEvent)->()) {
            self.init()
            self.block = block
        }
        override public func onSync(event:SyncEvent) {
            block(event)
        }
    }
    private var syncCallback:MACVSyncCallback?

    
    var dcd: DirectlyConnectedDevice?
    private var virtualMAC: VirtualDevice?
    // (id:, msg:, error:)
    private var appCallback:(String?,String?,String?)->() = { _,_,_ in }

    private var camera:Camera? = nil
    @objc let captureInterval = 5.0
    @objc var imageIndex = 0
    @objc var capturePhotoTimer:Timer? = nil
    
    override
    public init() {
        super.init()
        syncCallback = MACVSyncCallback(block: { (event) in
            //print("### In onSync callback ####")
            
            let virtualDevice:VirtualDevice = event.getVirtualDevice() as! VirtualDevice
            let storageObject:StorageObject = event.getSource()
            let format = "%@ : %@ : onSync : %@ = %@"
            var msg = String(format: format,
                             ISO8601DateFormatter().string(from: Date()),
                             virtualDevice.getEndpointId(),
                             storageObject.getName(),
                             "\(storageObject.getSyncStatus())")
            
            if storageObject.getSyncStatus() == SyncStatus.inSync {
                msg += " (\(storageObject.getLength())) bytes"
            }
            self.appCallback(nil, msg, nil)
        })
    }

    @objc convenience
    public init(appCallback: @escaping (String?,String?,String?)->()) {
        self.init()
        self.appCallback = { id, msg, error in
            if let msg = msg {
                print(msg)
            }
            if let error = error {
                print(error)
            }
            appCallback(id, msg, error) }
    }
    
    deinit {
        do {
            try dcd?.close()
        } catch {
            os_log("deinit(): Issue while closing Directly Connected Device.",
                   log: logger, type: .default)
        }
        dcd = nil
    }
    
    @objc public func start() throws -> String  {
        do {
            return try startDevice()
        } catch {
            os_log("start(): %@", log: logger, type:.default, "\(error)")
            throw error
        }
    }
    
    @objc public func stop() {
        capturePhotoTimer?.invalidate()
        stopDevice()
        imageIndex = 0
    }

    private func startDevice() throws -> String {

        dcd = try DirectlyConnectedDevice(path: Provisioning.getProvisioningFilePath(),
                                          password: Provisioning.getProvisioningPassword())
        
        // TODO: What is this for ?
        let persistentData = UserDefaults.standard
        if let hardwareId = persistentData.string(forKey: self.dcd!.getEndpointId()) {
            camera = Camera(id: hardwareId)
        } else {
            camera = Camera(id: UUID().uuidString)
            persistentData.set(camera?.hardwareId,
                               forKey: self.dcd!.getEndpointId())
        }
        do {
            if !dcd!.isActivated() {
                try self.dcd!.activate(deviceModels: [MOTION_ACTIVATED_CAMERA_MODEL_URN],
                                       callback: { endpointId, error in
                    if let error = error {
                        // TODO: show a GUI alert popup to exit, since can't do anything
                        // from this point
                        self.errorCallback(error:error)
                    } else {
                        // TODO: call startPhotoCapture directly
                        self.createVirtualDevice(endpointId: endpointId!,
                                            callback: self.startPhotoCapture)
                    }
                })

            } else {
                // TODO: call startPhotoCapture directly
                createVirtualDevice(endpointId: dcd!.getEndpointId(),
                                    callback: self.startPhotoCapture)
            }
        } catch {
            self.errorCallback(error:error as! ClientError)
            throw error
        }
        return dcd!.getEndpointId()
    }
    
    private func stopDevice() {
        do {
            try dcd?.close()
        } catch {
            os_log("stopDevice(): Issue while closing Directly Connected Device.",
                   log: logger, type: .default)
        }
        dcd = nil
    }
    
    func createVirtualDevice(endpointId:String, callback: @escaping () -> ())  {
        do {
            try self.dcd!.getDeviceModel(deviceModelUrn: MOTION_ACTIVATED_CAMERA_MODEL_URN,
                                         callback: { (deviceModel, error) in
                if let error = error {
                    self.errorCallback(error:error)
                    return
                }
                do {
                    self.virtualMAC =
                        try self.dcd!.createVirtualDevice(deviceId: endpointId,
                                                    deviceModel: deviceModel!)
                    self.appCallback(self.dcd?.getEndpointId(), nil, nil)
                    self.virtualMAC?.setOnError(callback: self.onError)
                    try self.virtualMAC?.setOnAction(actionName: "record",
                                                     callback: self.record)
                    try self.virtualMAC?.setOnAction(actionName: "scheduleRecording",
                                                     callback: self.scheduleRecording)
                    
                } catch  {
                    self.errorCallback(error:error as! ClientError)
                    return
                }
                // Run the camera in a separate thread not in the the current
                // call satck or getDeviceModel
                DispatchQueue(label: "RunCamera Queue", qos: .utility).async(
                    execute: { callback() })
            })
        } catch {
            self.errorCallback(error:error as! ClientError)
        }
    }
    
    let onErrorFormat:String = "%@ : onError : %@ : \"%@\""
    
    // TODO: Until there are error callbacks from the app.
    func errorCallback(error:ClientError) {
        let endpointId:String = self.virtualMAC?.getEndpointId() ?? "NA"
        self.appCallback(nil, nil, String(format:onErrorFormat,
            ISO8601DateFormatter().string(from: Date()),
            endpointId,
            "\(error)"))
    }

    func onError(event:ErrorEvent) {
        let device:VirtualDevice =  event.getVirtualDevice()
        self.appCallback(nil, String(format:onErrorFormat,
            ISO8601DateFormatter().string(from: Date()),
            device.getEndpointId(),
            event.getMessage()), nil)
    }

    private func getImagesDir() -> String {
        return (Bundle.main.url(forResource: "images",
                                withExtension: nil)?.path)!
    }
    
    private func getVideosDir() -> String {
        return (Bundle.main.url(forResource: "videos",
                                withExtension: nil)?.path)!
    }
    
    private func getFiles(dir:String) -> [String] {
        do {
            let dircontents = try FileManager.default.contentsOfDirectory(atPath: dir)
            let sortedArray = dircontents.sorted {
                $0.compare($1, options: .numeric) == .orderedAscending
            }
            return sortedArray
        } catch {
            return []
        }
    }
    
    private func startPhotoCapture() {
        imageIndex = 0
        // Loop to capture photos and update dewvice
        DispatchQueue.main.async(execute: {
            self.capturePhotoTimer =
                Timer.scheduledTimer(timeInterval: self.captureInterval,
                    target: self,
                    selector: #selector(MotionActivatedCamera.capturePhoto),
                    userInfo: nil, repeats: true)
        })
    }
    
    @objc
    private func capturePhoto() {
        let sonameformat = "%@"
        let msgformat = "%@ : Set : \"image\"=%@"
        
        let dir = getImagesDir()
        let images = getFiles(dir: dir)
        if images.isEmpty || images.count < imageIndex % images.count {
            print ("no images")
            return
        }
        var imageName = images[imageIndex % images.count]
        imageIndex += 1
        // Sort of assumes the images are in the "imageIndex" order
        let imagePath = dir + "/" + imageName
            
        // Make the top level file name unique with the endpoint id when
        // not adding the prefix
        if UNIQUE_FILENAME {
            imageName = (virtualMAC?.getEndpointId() ?? "id_unknown") + "_" + imageName
        }

        // May not want full path, probably just file name.
        let msg = String(format: msgformat,
                         formatDate(date:Date()), imagePath)

        do {
            /*
             * The name of video will be automatically
             * prefixed with the device's client id and "/"
             */
            let storageObject:StorageObject = try dcd!.createStorageObject(
                    name: String(format: sonameformat, imageName),
                    contentType:"image/jpeg")
            
            storageObject.setOnSync(callback: syncCallback)
            try storageObject.setInputPath(inputPath: imagePath)
            
            // TODO: Need to resolve issues with DATETIME attributes
            // It doesn't appear to be consistent throughout the library.
            // In this case it needs to "NSNumber"
            try virtualMAC?.update()
                .set(attributeName: IMAGE_ATTRIBUTE,
                     attributeValue: storageObject)
                .set(attributeName: IMAGE_TIME_ATTRIBUTE,
                     attributeValue: Int(Date().timeIntervalSince1970 * 1000.0) as AnyObject)
                .finish();
            //print(msg)
            self.appCallback(nil, msg, nil)
        } catch {
            self.appCallback(nil, "\(error)", nil)
        }
    }

    class VideoUploader {
        
        let dcd:DirectlyConnectedDevice
        let virtualDevice:VirtualDevice
        let videoFile:String
        let videoName:String
        let duration:Int
        var startTime:Int?
        let syncCallback:MACVSyncCallback
        let appCallback:(String?,String?,String?)->()
    
        init(dcd:DirectlyConnectedDevice, virtualDevice:VirtualDevice,
                videoFile:String, videoName:String, duration:Int, startTime:Int?=nil,
                syncCallback:MACVSyncCallback,
                appCallback: @escaping (String?,String?,String?)->()) {
            self.dcd = dcd
            self.virtualDevice = virtualDevice
            self.videoFile = videoFile
            self.videoName = videoName
            self.duration = duration
            self.startTime = startTime
            self.syncCallback = syncCallback
            self.appCallback = appCallback
        }
        
        func start() {
            let format = "%@"
            var deadline:DispatchTime = DispatchTime.now()
            
            // startTime may be earlier than "now" just log that fact if it
            // happens and record anyway
            if startTime != nil {
                let startDate:Date = Date(timeIntervalSince1970: Double(startTime!) / 1000)
                // seconds
                let delay:Int = (startDate.timeIntervalSinceNow as NSNumber).intValue
                if delay < 0 {
                    if abs(Int32(delay)) > duration {
                        self.appCallback(nil,
                            String(format:"VideoUploader.start(): startTime is before \"now\" by %d seconds, and the difference is greater than the duration %d seconds, scheduling recording anyway.",
                                     abs(Int32(delay)), duration), nil)
                        
                    } else {
                        self.appCallback(nil,
                            String(format:"VideoUploader.start(): startTime is before \"now\" by %d seconds scheduling recording anyway.",
                                     delay), nil)
                    }
                } else {
                    deadline = deadline + .seconds(delay) + .seconds(duration)
                }
            } else {
                self.startTime = ((Date().timeIntervalSince1970 * 1000.0) as NSNumber).intValue
            }
            
            
            // Simulate recording for duration
            DispatchQueue(label: "recording queue", qos: .utility).asyncAfter(
                deadline: deadline,
                execute: DispatchWorkItem(qos: .utility, block: {
                    do {
                        let storageObject:StorageObject  = try (self.dcd.createStorageObject(
                            name: String(format: format, self.videoName),
                            contentType: "video/mp4"))
                        
                        storageObject.setOnSync(callback: self.syncCallback)
                        try storageObject.setInputPath(inputPath: self.videoFile)
                        try storageObject.setCustomMetadata(key:"duration",
                                    value:(self.duration as NSNumber).stringValue)

                        let formattedDate:String = formatDate(date: Date(
                                timeIntervalSince1970: Double(self.startTime!) / 1000.0))
                        
                        let msg = String(format: "%@ : Set : recording = %@",
                                         formattedDate,
                                         storageObject.getURI())
                        
                        let DATA_FAILED_FORMAT = "%@ : onError : "
                        
                        // A callback for receiving an error if the record data
                        // message fails.
                        func dataFailed(event:ErrorEvent) {
                            var msg:String = String(format:DATA_FAILED_FORMAT,
                                                    formatDate(date:Date()))
                            
                            var namedValue:NamedValue? = event.getNamedValue()
                            var comma:String = ""
                            while namedValue != nil {
                                let attribute:String = namedValue!.getName()
                                let value:Any = namedValue!.getValue() ?? "nil"
                                msg = msg + comma + "\"" + attribute + "\"=\(value)"
                                namedValue = namedValue!.next()
                                comma = ","
                            }
                            self.appCallback(nil, msg, nil)
                        }
                        
                        let recording:DeviceLib.Data = try self.virtualDevice.createData(
                            format: MOTION_ACTIVATED_CAMERA_MODEL_URN + ":recording");
                        recording.setOnError(callback:dataFailed)
                        try recording.set(field: "video", value: storageObject)
                            .set(field: "startTime", value: self.startTime as AnyObject)
                            .set(field: "duration", value: self.duration as AnyObject)
                            .submit()
                        
                        self.appCallback(nil, msg, nil)
                        
                    } catch {
                        self.appCallback(nil, "\(error)", nil)
                    }
                }))
        }
    }
    
    // JavaSE sample
    // assumes videos.length > 0 and videos are 15 second increments.
    // the video files are indexed by interval 0-15, 16-31,
    // ... videos[Math.min(duration/15-1,videos.length-1)];
    
    private func record(actionEvent:ActionEvent) {
        let virtualDevice:VirtualDevice = actionEvent.getVirtualDevice()
        let arguments:NamedValue = actionEvent.getNamedValue()
        let d:Int = (arguments.getValue() as! NSNumber).intValue
        
        guard d > 0 else {
            self.appCallback(nil, "record(): duration is not > 0, action not performed", nil)
            return
        }
        
        // round to nearest increment of 15
        let duration = (d + 14) / 15 * 15
        // assumes videos.length > 0 and videos are 15 second increments.
        let dir = getVideosDir()
        let videos = getFiles(dir: dir)
        var videoName = videos[min(duration/15-1, videos.count-1)]
        let videoFile = dir + "/" + videoName
        
        let startDate:Date = Date()
        let msg = String(format:"%@ : Call : record : %d seconds",
                         formatDate(date:startDate), d)
        self.appCallback(nil, msg, nil)
        
        // Make the top level file name unique with the endpoint id when
        // not adding the prefix
        if UNIQUE_FILENAME {
            videoName = virtualDevice.getEndpointId() + "_" + videoName
        }
        
        let vu:VideoUploader = VideoUploader(dcd:self.dcd!,
                                             virtualDevice:virtualDevice,
                                             videoFile:videoFile,
                                             videoName:videoName,
                                             duration:duration,
                                             startTime: nil,
                                             syncCallback:self.syncCallback!,
                                             appCallback:self.appCallback)
        vu.start()
    }
    
    private func scheduleRecording(actionEvent:ActionEvent) {
        let virtualDevice:VirtualDevice = actionEvent.getVirtualDevice()
        let arguments:NamedValue = actionEvent.getNamedValue()
        var nv:NamedValue? = arguments
        var duration:Int = 0
        var startTime:Int = 0
        while nv != nil {
            switch (nv!.getName()) {
            case "duration":
                duration = (nv!.getValue() as! NSNumber).intValue
                break
            case "startTime":
                startTime = (nv!.getValue() as! NSNumber).intValue
                break
            default:
                break
            }
            nv = nv!.next()
        }
        
        guard duration > 0 else {
            self.appCallback(nil, "scheduleRecording(): duration is not > 0, action not performed", nil)
            return
        }
        
        // round to nearest increment of 15
        duration = (duration + 14) / 15 * 15
        // assumes videos.length > 0 and videos are 15 second increments.
        let dir = getVideosDir()
        let videos = getFiles(dir: dir)
        var videoName = videos[min(duration/15-1, videos.count-1)]
        let videoFile = dir + "/" + videoName

        //let msg = String(format:"%@ : Call : record : %d seconds",
        // formatDate(date:startDate), duration)
        //print(msg)
        //self.updateGUI(nil, msg)
        
        // Make the top level file name unique with the endpoint id when
        // not adding the prefix
        if UNIQUE_FILENAME {
            videoName = virtualDevice.getEndpointId() + "_" + videoName
        }
        
        let vu:VideoUploader = VideoUploader(dcd:self.dcd!,
                                             virtualDevice:virtualDevice,
                                             videoFile:videoFile,
                                             videoName:videoName,
                                             duration:duration,
                                             startTime: startTime > 0 ? startTime : nil,
                                             syncCallback:self.syncCallback!,
                                             appCallback:self.appCallback)
        vu.start()
    }
}
