/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit

class MotionActivatedCameraView: UIViewController, UITextViewDelegate {
    
    @objc var motionActivatedCamera:MotionActivatedCamera?
    
    @IBOutlet var motionActivatedViewOutlet: UIView!
    @IBOutlet weak var idValueOutlet: UILabel!
    @IBOutlet weak var outputTextViewOutlet: UITextView!
    
    @IBOutlet weak var stopButtonOutlet: UIButton!

    override func viewDidLoad() {
        super.viewDidLoad()
        outputTextViewOutlet.delegate = self
        
        
        // Set up notifications for terminate.
        // The app will shut down on either of these events.
        // This works in conjunction with the Info.plist property
        // Application does not run in background, or UIApplicationExitsOnSuspend.
        // When the "STOP" button is pressed, the app is put into suspend, and
        // because of the property "applicationWillTerminate" is called.
        NotificationCenter.default.addObserver(self, selector: #selector(self.terminate),
                                               name: UIApplication.willTerminateNotification, object: nil)
    }
        
    override func viewWillAppear(_ animated:Bool) {
        super.viewWillAppear(animated)
        do {
            // Should this be created on every load ?
            self.motionActivatedCamera =
                MotionActivatedCamera(appCallback: self.appCallback)
            // Returns endpointId
            self.idValueOutlet?.text =
                try self.motionActivatedCamera?.start()
        } catch {
            showError(controller: self,
                      alertTitle: "Motion Activated Camera",
                      message: "Could not start Motion Activated Camera",
                      actionTitle: "Exit",
                      handler: { action in
                        self.outputTextViewOutlet.text = ""
                        // When the app is suspended, because of the Info.plist property
                        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
                        // observer is called and the "terminate" method will be called.
                        UIControl().sendAction(#selector(NSXPCConnection.suspend),
                                               to: UIApplication.shared, for: nil)
            })
        }

        self.idValueOutlet?.text =
        motionActivatedCamera?.dcd?.getEndpointId()
        
        if Provisioning.isProvisioningStatic() {
            self.navigationItem.setHidesBackButton(true, animated: false)
        }
    }
    
    override func viewWillDisappear(_ animated:Bool) {
        super.viewWillDisappear(animated)
        
        // back button clicked (Reset)
        if isMovingFromParent {
            Provisioning.reProvision()
            navigationController?.popToRootViewController(animated: false)
        }
    }
    
    // Called from AppDelegate, when application exits
    @objc func close() {
        self.motionActivatedCamera?.stop()
    }
    
    @objc func appCallback(id:String?, msg:String?, error:String?) {
        if let error = error {
            showError(controller: self,
                      alertTitle: "Motion Activated Camera",
                      message: error,
                      actionTitle: "Exit",
                      handler: { action in
                        exit(0)
            })
        } else {
            self.updateGUI(id:id, text:msg)
        }
    }
    
    @objc func updateGUI(id: String? = nil, text: String? = nil) {
        DispatchQueue.main.async(execute: {
            if let id = id {
                self.idValueOutlet?.text = id
            }
            if let text = text {
                self.outputTextViewOutlet.text = self.outputTextViewOutlet.text! + "\n" + text
                // Force text field to (invisibly) scroll until latest line is at bottom
                let bottom: NSRange =
                    NSMakeRange(self.outputTextViewOutlet.text.count - 10, 1)
                self.outputTextViewOutlet.scrollRangeToVisible((bottom))
            }
        })
    }
    
    @IBAction func stopButtonAction(_ sender: UIButton) {
        self.outputTextViewOutlet.text = ""
        // When the app is suspended, because of the Info.plist property
        // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
        // observer is called and the "terminate" method will be called.
        UIControl().sendAction(#selector(NSXPCConnection.suspend),
                               to: UIApplication.shared, for: nil)
    }
    
    @objc func terminate() {
        NotificationCenter.default.removeObserver(self)
        motionActivatedCamera?.stop()
        saveSettingsFile()
    }
}
