/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit

class ChooseProvisionerView: UIViewController {

    @IBOutlet weak var fileProvisionerButton: UIButton!
    @IBOutlet weak var networkProvisionerButton: UIButton!
    @IBOutlet weak var withPoliciesLabelOutlet: UILabel!
    @IBOutlet weak var withPoliciesSwitchOutlet: UISwitch!
    
    // Requires the application to create Main.storyboard
    // and an intial view controller
    @objc var MainStoryboard: UIStoryboard? = nil
    
    private func toggleWithPolicies(withPolicies:Bool) {
        if withPoliciesSwitchOutlet.isOn {
            withPoliciesLabelOutlet.text = "Use device policies"
        } else {
            withPoliciesLabelOutlet.text = "Do not use device policies"
        }
    }
    
    @IBAction func withPoliciesSwitchAction(_ sender: UISwitch) {
        // Set the the state from the withPolicesSwitch
        Provisioning.setWithPolicies(withPolicies:sender.isOn)
        toggleWithPolicies(withPolicies: Provisioning.getWithPolicies())
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        // Provisioning.getWithPolicies() uses the global WITH_POLCIIES
        // variable defined in Constants.swift. The default value is false.
        // This means that the Device code should choose a code path
        // that does not use policies.
        // See withPoliciesSwitchAction.
        withPoliciesSwitchOutlet.isOn = Provisioning.getWithPolicies()
        toggleWithPolicies(withPolicies: withPoliciesSwitchOutlet.isOn)
        self.title = NAVBARTITLE
    }

    var firstTime:Bool = true
    override func viewWillAppear(_ animated: Bool) {
        self.navigationItem.setHidesBackButton(true, animated: false)
        if let provError:String = Provisioning.getError() {
            showError(controller: self, alertTitle: "Provisioner Alert",
                      message: provError, actionTitle: "OK",
                      handler: {(action:UIAlertAction?) -> () in
                // When the app is suspended, because of the Info.plist property
                // UIApplicationWillExitOnSuspend the UIApplicationWillTerminate
                // observer is called and the "terminate" method will be called.
                UIControl().sendAction(#selector(NSXPCConnection.suspend),
                                           to: UIApplication.shared, for: nil)
            })
        }
        // There are a couple of cases.
        // - If the device was statically provisioned, then it cannot be reprovisioned
        //   do not show the provisioning choices. We should never get here then
        //   after the first time.
        // - If the device is not statically provisioned and has not been previously
        //   provisioned, show the provisioning choices.
        //   If previously provisioned move to the next screen. There will be a
        //   "cancel" back button that can send the user back here. In that case
        //   show the provisioning choices, firstTime == false. This is effectively a reprovision.
        let isStatic:Bool = Provisioning.isProvisioningStatic()
        let isProvisioned:Bool = Provisioning.isProvisioned()
        if (firstTime && !isStatic && !isProvisioned) || (!firstTime && !isStatic) {
            super.viewWillAppear(false)
            navigationItem.backBarButtonItem!.title = "Cancel"
        } else {
            // Have provisioning infomation go right to the App initial view
            // Every App that wants to use generic File and Network bootstrapping
            // must have a "Main" storyboard with at least one view controller
            // tagged as the "initial view controller"
            if self.MainStoryboard == nil {
                self.MainStoryboard = UIStoryboard(name: "Main", bundle: nil)
            }
            navigationItem.backBarButtonItem!.title = "Reset"
            performSegue(withIdentifier: "ProvisionToMain", sender: self)
        }
        if firstTime {
            firstTime = false
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(false)
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
}
