/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * A simulated humidity sensor for use in the samples. The sensor has
 * humidity and maximum threshold attributes.
 */
public class HumiditySensor {
    
    // The min range for values is 0% humidity
    private var hmin: Int = 0
    // The hmin range for values is 100% humidity
    private var hmax: Int = 100
    // Maximum humidity threshold, writable
    private var maxThreshold: Int = 80
    // A unique identifier for this sensor
    private var hardwareId:String = ""
    // The rotator angle in degrees
    private var angle:Int = 0

    private let humidityQueue:DispatchQueue =
        DispatchQueue(label:"HumiditySensor humidity queue", qos:.utility)

    /**
     * Create a HumiditySensor
     * - parameter id: a unique identifier for this sensor
     */
    public init(id: String) {
        hardwareId = id
    }
    
    /**
     * Get the current humidity value. The value returned is between 0 and 100,
     * representing a percent humidity.
     * - returns: the current humidity value.
     */
    public func getHumidity() -> Int {
        return humidityQueue.sync(execute: {
            let amplitude:Double = (Double(self.maxThreshold) - Double(self.hmin))/2.0
            let angleinradians:Measurement =
                Measurement(value: Double(self.angle),unit: UnitAngle.degrees)
                    .converted(to:.radians)
            let delta:Double = amplitude * sin(angleinradians.value)
            
            var humidity:Double = 0.0
            var result:Int = 0
            if self.angle == 90 {
                result = maxThreshold + 2
            } else {
                humidity = Double(hmin) + amplitude + delta
                result = Int(humidity.rounded())
            }
            self.angle = (self.angle + 15) % 360
            return clamp(val:result, minimum:hmin, maximum:hmax)
        })
    }
    
    /**
     * Set the maximum percent humidity threshold for alerts. The value is
     * clamped to the range [60..100] and represents a whole percentage.
     * @param threshold a value between 60 and 100
     */
    public func setMaxThreshold(threshold: Int) {
        if threshold < 60 {
            maxThreshold = 60
        } else if threshold > 100 {
            maxThreshold = 100
        } else {
            maxThreshold = threshold
        }
    }
    
    /**
     * Get the maximum threshold value.
     * - returns: the maximum threshold
     */
    public func getMaxThreshold() -> Int {
        return self.maxThreshold
    }
    
    /**
     * Get the manufacturer name, which can be used as part of the
     * device meta-data.
     * - returns: the manufacturer name
     */
    public func getManufacturer() -> String {
        return "Sample"
    }
    
    /**
     * Get the model number, which can be used as part of the
     * device meta-data.
     * - returns: the model number
     */
    public func getModelNumber() -> String {
        return "MN-" + self.hardwareId
    }
    
    /**
     * Get the serial number, which can be used as part of the
     * device meta-data.
     * - returns: the serial number
     */
    public func getSerialNumber() -> String {
        return "SN-" + self.hardwareId
    }
    
    /**
     * Get the hardware id, which can be used as part of the
     * device meta-data.
     * - returns: the hardware id
     */
    public func getHardwareId() -> String {
        return self.hardwareId
    }
    /**
     * Clamp value between min and max, inclusive.
     * - parameter val: the value
     * - parameter minimum: the minimum allowable value
     * - parameter maximum: the maximum allowable value
     * - return: A value such that {@code min <= value <= max}
     */
    private func clamp(val:Int, minimum:Int, maximum:Int) -> Int {
        return max(minimum, min(maximum, val));
    }
}
