/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit
import Foundation
import CFNetwork
import IPTools

class NetworkProvisioner {

    let DISCOVER_REQUEST:UInt8 = 0x01
    let DISCOVER_RESPONSE:UInt8 = 0x02
    let PROVISION_REQUEST:UInt8 = 0x03
    let PROVISION_RESPONSE:UInt8 = 0x04

    static let SUCCESS: UInt8 = 0x00
    static let FAILURE: UInt8 = 0x01
    static let INIT:UInt8 = 0xFF

    var START_PROVISION:Bool = false

    enum ProvisionError:Error {
        case fail(String)
    }

    var provisionSocket:Int32 = 0
    var provisionHost:String = "238.163.7.96"
    var provisionPort:Int = 4456
    var socketBufSize = 65535
    
    var dest_sa:sockaddr? = nil
    var taStore:String
    var taStorePassword:String
    
    /**
     * NetworkProvisioner calls this handler when it has received a response from a
     * request. It is not called for all NetworkProvisioner status changes. It is
     * intended to indicate that the file has been downloaded successfully or
     * unsuccessfully.
     */
    private var provisionFileCallback:((UInt8,ProvisionError?) -> ())

    init?(taStore:String, taStorePassword:String,
            provisionFileCallback:@escaping (UInt8,ProvisionError?)->()) {
        
        self.taStore = taStore
        self.taStorePassword = taStorePassword
        self.provisionFileCallback = provisionFileCallback
        
        provisionSocket = Darwin.socket(AF_INET, SOCK_DGRAM, 0)
        guard provisionSocket >= 0 else {
            self.provisionFileCallback(NetworkProvisioner.INIT,
                ProvisionError.fail("provisioner_receive(): Darwin.socket failed with \(provisionSocket)"))
            return nil
        }
        DispatchQueue(label:"NetworkProvisioner Queue", qos:.userInitiated).async() {
            do {
                try self.setSocketOptions()
                let dataAndDest:(data:[UInt8], dest:sockaddr) = try self.provisioner_receive()
                self.dest_sa = dataAndDest.dest
                try self.provision(data:dataAndDest.data, dest:dataAndDest.dest)
            } catch {
                self.provisionFileCallback(NetworkProvisioner.INIT,
                    error as? ProvisionError ?? ProvisionError.fail("\(error)"))
            }
        }
    }

    func setSocketOptions() throws {

        // We want blocking
        //guard fcntl(self.provisionSocket, F_SETFL, O_NONBLOCK) != -1 else {
        //    throw ProvisionError.fail("fnctl: O_NONBLOCK failed error \(errno)")
        //}

        for socketoption in [
                (SO_REUSEADDR, "SO_REUSEADDR", 1),
                (SO_NOSIGPIPE, "SO_NOSIGPIPE", 1),
                (SO_SNDBUF, "SO_SNDBUF", socketBufSize),
                (SO_RCVBUF, "SO_RCVBUF", socketBufSize),
                (SO_BROADCAST, "SO_BROADCAST", 1)] {
            var value:socklen_t = socklen_t(socketoption.2)
            guard setsockopt(self.provisionSocket, SOL_SOCKET, socketoption.0,
                             &value, socklen_t(MemoryLayout<socklen_t>.size)) != -1 else {
                throw ProvisionError.fail("setsockopt: \(socketoption.2) failed errno \(errno)")
            }
        }
    }

    func provisioner_receive() throws -> (data:[UInt8], dest:sockaddr/*, host:String, port:Int*/) {
        
        var status:Int32 = 0
        
        var provision_sa_in:sockaddr_in = sockaddr_in()
        provision_sa_in.sin_len = __uint8_t(MemoryLayout<sockaddr_in>.size)
        provision_sa_in.sin_family = sa_family_t(AF_INET)
        provision_sa_in.sin_port = in_port_t(provisionPort).bigEndian
        provision_sa_in.sin_addr.s_addr = in_addr_t(INADDR_ANY.bigEndian)

        // cast sockaddr_in to sockaddr
        status = withUnsafePointer(to: &provision_sa_in, { ptr in
            ptr.withMemoryRebound(to: sockaddr.self, capacity: 1, { mem in
                Darwin.bind(provisionSocket, mem, socklen_t(MemoryLayout<sockaddr_in>.size))
            })
        })
        guard status == 0 else {
            self.stop()
            throw ProvisionError.fail("provisioner_receive(): Darwin.bind failed with \(status) errno \(errno)")
        }

        let portStr:String = String(format:"%hu", 0)
        var provisionhost_ai_hints = addrinfo(
                ai_flags:AI_NUMERICHOST,
                ai_family:PF_UNSPEC,
                ai_socktype: SOCK_DGRAM,
                ai_protocol: IPPROTO_UDP,
                ai_addrlen: 0,
                ai_canonname: nil,
                ai_addr: nil,
                ai_next: nil
        )
        var provisionhost_ai:UnsafeMutablePointer<addrinfo>? = nil
        status = getaddrinfo(provisionHost, portStr,
                             &provisionhost_ai_hints, &provisionhost_ai)
        guard status == 0 else {
            self.stop()
            throw ProvisionError.fail("provisioner_receive(): getaddrinfo for \(provisionHost) failed with \(status) errno \(errno)")
        }
        
        var provisiongroup_sa:sockaddr? = nil
        var have_sa:Bool = false
        var next_ai:addrinfo? = provisionhost_ai?.pointee
        while let ai:addrinfo = next_ai {
            if !have_sa  && ai.ai_family == AF_INET {
                provisiongroup_sa = ai.ai_addr?.pointee
                have_sa = true
            }
            next_ai = ai.ai_next?.pointee
        }
        freeaddrinfo(provisionhost_ai)

        // cast sockaddr to sockaddr_in
        status = withUnsafePointer(to: &provisiongroup_sa!, { ptr in
            ptr.withMemoryRebound(to: sockaddr_in.self, capacity: 1, {mem in
                var imreq:ip_mreq = ip_mreq(
                    imr_multiaddr:mem.pointee.sin_addr,
                    imr_interface:provision_sa_in.sin_addr
                )
                return setsockopt(provisionSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                                  &imreq, socklen_t(MemoryLayout<ip_mreq>.size))
            })
        })
        guard status == 0 else {
            self.stop()
            throw ProvisionError.fail("provisioner_receive(): setsockopt() IP_ADD_MEMBERSHIP failed with \(status) errno \(errno)")
        }

        var receive_sa:sockaddr = sockaddr()
        var receive_sa_len:socklen_t = socklen_t(MemoryLayout<sockaddr_in>.size)

        let receive_bufsize:size_t = socketBufSize
        var receive_buf:[UInt8] = Array<UInt8>(repeating: 0, count: receive_bufsize)

        let received_len:Int = recvfrom(provisionSocket, &receive_buf, receive_bufsize, 0,
                                   &receive_sa, &receive_sa_len)
        guard received_len > 0 else {
            if received_len < 0 {
                throw ProvisionError.fail("provisioner_receive(): recvfrom() returned \(received_len) errno = \(errno)")
            } else {
                throw ProvisionError.fail("provisioner_receive(): recvfrom() returned \(received_len) bytes")
            }
        }
        // Get only the returned bytes not the whole buffer
        receive_buf = Array<UInt8>(receive_buf[0..<received_len])
        
    //    var host_buf:[Int8] = Array<Int8>(repeating: 0, count: Int(INET_ADDRSTRLEN))
    //    var host_port:Int? = nil
    //    let host:String? = withUnsafePointer(to: &receive_sa, { ptr in
    //        ptr.withMemoryRebound(to: sockaddr_in.self, capacity: 1, {mem in
    //            var sin_addr:in_addr = mem.pointee.sin_addr
    //            if let _host:UnsafePointer<Int8> = inet_ntop(AF_INET, &sin_addr,
    //                                                         &host_buf, socklen_t(host_buf.count)) {
    //                host_port = Int(mem.pointee.sin_port)
    //                return String(data: Foundation.Data(bytes: _host,
    //                                                    count: Int(INET_ADDRSTRLEN)),
    //                              encoding:.utf8)
    //            } else {
    //                return nil
    //            }
    //        })
    //    })
        
    //    guard host != nil , host_port != nil else {
    //        close(PROVISIONSOCKET)
    //        throw ProvisionError.fail("provisioner_receive(): Could not resolve multicast \(PROVISIONHOST) host and port")
    //    }
        return (data:receive_buf, dest:receive_sa/*, host:host!, port:host_port!*/)
    }

    private func provisioner_send(data:[UInt8], dest:sockaddr) throws {
        var _dest:sockaddr = dest
        let result:Int = sendto(provisionSocket, data, data.count, 0,
                                &_dest, socklen_t(MemoryLayout<sockaddr>.size))
        guard result > 0 else {
            self.stop()
            if result < 0 {
                throw ProvisionError.fail("provisioner_send(): failed with \(result) errno \(errno)")
            } else {
                throw ProvisionError.fail("provisioner_send(): failed no bytes sent")
            }
        }
    }

    private func isProvisioned() -> Bool {
        return FileManager.default.fileExists(atPath: self.taStore) &&
            validateTam(path: self.taStore, password: self.taStorePassword)
    }

    private func provisionResponse(status:UInt8) -> [UInt8] {
        var response: [UInt8] = []
        response.append(PROVISION_RESPONSE)
        response.append(status)
        return response
    }

    private func updateStore(buf:[UInt8], offset:Int) throws -> String {
        var length:Int = buf.count
        length -= offset
        var newoffset:Int = offset
        
        // Need to move past the request length
        var reqlen:Int = Int(buf[newoffset] & 0xFF) << 8
        newoffset += 1
        length -= 1
        reqlen += Int(buf[newoffset] & 0xFF)
        newoffset += 1
        length -= 1

        guard length == reqlen else {
            self.stop()
            throw ProvisionError.fail("Provisioning information too short")
        }
        
        let subbuf = buf[newoffset...reqlen]
        let data = Data(subbuf)
        let tampath:String = self.taStore

        guard FileManager.default.createFile(atPath: tampath, contents: data),
                FileManager.default.fileExists(atPath: tampath) else {
            self.stop()
            throw ProvisionError.fail("updateStore(): failed to write \(tampath)")
        }
        return tampath
    }

    private func provision(data: [UInt8], dest:sockaddr) throws {

        switch data[0] {
        case DISCOVER_REQUEST:

            let key:String = "MAC"
            let value:String? = UIDevice.current.identifierForVendor?.uuidString
            
            var sendData: [UInt8] = []
            sendData.append(DISCOVER_RESPONSE)
            var offset = 1
            
            var bytes = Array(key.utf8)
            sendData.append(UInt8(bytes.count))
            sendData.append(contentsOf: bytes)
            offset += bytes.count + 1
            
            bytes = Array(value!.utf8)
            sendData.append(UInt8(bytes.count))
            sendData.append(contentsOf: bytes)
            offset += bytes.count + 1
            
            try provisioner_send(data:sendData, dest:dest)
            break
            
        case PROVISION_REQUEST:
            if START_PROVISION == false {
                START_PROVISION = true
                
                // Check for minimum possible valid length
                if data.count < 5 {
                    print("\tProvisioning was unsuccessful, format was not correct.\n")
                    let response:[UInt8] = provisionResponse(status:NetworkProvisioner.FAILURE)
                    try provisioner_send(data:response, dest:dest)
                    self.provisionFileCallback(0x01, nil)
                    return
                }
                
                var responseStatus:UInt8 = NetworkProvisioner.FAILURE
                var storeStatus:UInt8 = 0x01
                do {
                    let _ = try updateStore(buf:data, offset:1)
                    if isProvisioned() {
                        print("\tSuccessfully provisioned.\n")
                        responseStatus = NetworkProvisioner.SUCCESS
                        storeStatus = 0x02
                    } else {
                        print("\tProvisioning was unsuccessful.\n")
                        storeStatus = 0x03
                    }
                } catch {
                    print("Error \(error)")
                    storeStatus = 0x01
                }
                // Respond
                let response:[UInt8] = provisionResponse(status:responseStatus)
                try provisioner_send(data:response, dest:dest)
                self.provisionFileCallback(storeStatus,nil)
            }
            break
            
        default:
            break
        }
    }

    // Called by the application to indicate provisioning failed
    public func cancel() {
        do {
            try provisioner_send(data: provisionResponse(status:NetworkProvisioner.FAILURE), dest:sockaddr())
        } catch {
            print("cancel(): \(error)")
        }
        // Don't call the callback for this since the callback is only
        // until file is downloaded
        close(provisionSocket)
    }
    
    public func stop() {
        close(provisionSocket)
    }
}
