/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import UIKit

// The filename for the Network Provisioned file
fileprivate let kNetProvFile = "networkProvisioner.conf"
fileprivate var networkProvisioner:NetworkProvisioner?
fileprivate var provisioning:Bool = false

fileprivate func networkProvisionerStop() {
    networkProvisioner?.stop()
    provisioning = false
    networkProvisioner = nil
}

class NetworkProvisioningPasswordView: PasswordView {

    // Requires the application to create Main.storyboard
    // and an intial view controller
    private var MainStoryboard: UIStoryboard? = nil
    

    @IBOutlet weak var activityOutlet: UIActivityIndicatorView! {
        didSet {
            activityOutlet.style = .whiteLarge
            let transform: CGAffineTransform =
                CGAffineTransform(scaleX: 4.0, y: 4.0)
            activityOutlet.transform = transform
            activityOutlet.color = UIColor.black
            activityOutlet.hidesWhenStopped = true
        }
    }

    override func viewDidLoad() {
        if MainStoryboard == nil {
            MainStoryboard = UIStoryboard(name: "Main", bundle: nil)
        }
        super.viewDidLoad()
        self.title = NAVBARTITLE
    }
    
    override func viewWillAppear(_ animated: Bool) {
        if MainStoryboard == nil {
            MainStoryboard = UIStoryboard(name: "Main", bundle: nil)
        }
        passwordOutlet.isEnabled = true
        super.viewWillAppear(false)
    }
    
    // Only two ways to leave the password screen
    // Clicking keyboard Continue or nav bar Cancel
    override func viewWillDisappear(_ animated:Bool) {
        if provisioning {
            networkProvisioner?.cancel()
        }
        networkProvisionerStop()
        super.viewWillDisappear(false)
    }

    override func verifyPassword(password: String) -> Bool {
        return true
    }
    
    override func handleContinue() {
        networkProvision()
    }
    
    @objc func successView() {
        networkProvisionerStop()
        let path = FileParser.sharedInstance.documentsURL().appendingPathComponent(
            kNetProvFile, isDirectory: false)
        Provisioning.setProvisioningFilePath(pf: path.path)
        Provisioning.setProvisioningPassword(pp: getPassword())
        saveSettingsFile()
        
        // Every App that wants to use generic File and Network provisioning
        // must have an "Main" storyboard with at one view controller
        // tagged as the "initial view controller"
        if self.activityOutlet.isAnimating {
            self.activityOutlet.stopAnimating()
        }
        // Continue keyboard button pressed
        performSegue(withIdentifier: "ProvisionToMain", sender: self)
    }
    
    @objc func failView(_ action:UIAlertAction) {
        if self.activityOutlet.isAnimating {
            self.activityOutlet.stopAnimating()
        }
        let _ = self.navigationController?.popToRootViewController(animated: false)
        networkProvisionerStop()
    }
    
    @objc func networkProvision() {
        self.activityOutlet.startAnimating()
        DispatchQueue.main.async(execute: {
            self.networkProvisioning(filename: getDocumentPath(file: kNetProvFile),
                                password: self.getPassword(),
                                failView: self.failView,
                                successView: self.successView)
        })
    }

    @objc func networkProvisioning(filename: String, password: String,
                             failView: @escaping ((UIAlertAction) -> ()),
                             successView: @escaping (() -> ()))  {
        
        provisioning = true
        networkProvisioner = NetworkProvisioner(taStore: filename,
                                                taStorePassword: password,
                                                provisionFileCallback: { (status,error) in
            if let error = error {
                showError(controller: self, alertTitle: "Provisioning Alert",
                          message: "Cannot create NetworkProvisioner \(error)",
                          actionTitle: "OK",
                          handler: failView)
                return
            }
            switch status {
            // Bad format or an exception occurred
            case 0x01:
                showError(controller: self, alertTitle: "Provisioning Alert",
                          message: "Provisioning was unsuccessful, format was not correct.",
                          actionTitle: "OK",
                          handler: failView)
                break
            // File downloaded and validated
            case 0x02:
                if !FileManager.default.fileExists(atPath: filename) {
                    // Shouldn't happen since this was checked in the networkProvisioner
                    //networkProvisioner?.notProvisioned()
                    showError(controller: self, alertTitle: "Provisioning Alert",
                              message: "The \(filename) file does not exist.",
                        actionTitle: "OK",
                        handler: failView)
                    return
                }
                
                // TODO: pop this view off the navigator stack so that the
                // password screen if cancelled will got directly to the
                // chooser view
                DispatchQueue.main.async(execute:{
                    successView()
                })
                break
            // File does not exist
            case 0x03:
                showError(controller: self,alertTitle: "Provisioning Alert",
                          message: "Provisioning was unsuccessful.",
                          actionTitle: "OK",
                          handler: failView)
                break
            // Don't see this state in the Bootstrapper.
            case 0x04:
                showError(controller: self,alertTitle: "Provisioning Alert",
                          message: "Error during getting data from truststore.",
                          actionTitle: "OK",
                          handler: failView)
                break
                
            default:
                break
            }
        })
    }
}
