/*
 *  Copyright © 2016, 2019, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * A simulated temperature sensor for use in the samples. The sensor has
 * temperature, maximum temperature threshold, and minimum temperature
 * threshold attributes.
 */
public class TemperatureSensor {

    // Whether the sensor is powered on (true), or off (false)
    private var power:Bool = false
    // The time (measured in EPOCH) at which the system was powered ON or reset
    private var startTime:Int64 = 0
    // The minimum value measured by the sensor since power ON or reset
    
    private var minTemp:Double = .greatestFiniteMagnitude
    // The maximum value measured by the sensor since power ON or reset
    private var maxTemp:Double = .leastNormalMagnitude
    // The minimum threshold is in the range [-20,0]
    private var minThreshold:Int = 0
    // The maximum threshold is in the range [65,80]
    private var maxThreshold:Int = 70
    // A unique identifier for this sensor
    private var hardwareId:String = ""
    // The current rotator angle in degrees
    private var angle:Int = 0
    
    // The maximum temperature this sensor can record, according to the device model.
    private let UPPER_TEMP_LIMIT:Double = 80.0
    
    // The minimum temperature this sensor can record, according to the device model.
    private let LOWER_TEMP_LIMIT:Double = -20.0
    
    private let temperatureQueue:DispatchQueue =
        DispatchQueue(label:"TemperatureSensor temperature queue",
                      qos:.utility)
    
    // The TemperatureSensor simulates temperature data by generating values that
    // fluctuate around the 'set point', arbitrarily set to 65 Celsius Degrees.
    private var setPoint: Double = 65.0
    // amplitude is the maximum deviation from the 'set point'

    private var amplitude:Double = 0.0
    // the simulated data
    private var temperatures:[Double]?
    // index into temperature data
    private var index:Int = 0
    
    /**
     * Create a TemperatureSensor
     * - parameter id: a unique identifier for this sensor
     */
    public init(id: String) {
        hardwareId = id
        power(on: true)
    }
    
    /**
     * Set the power on or off.
     * - parameter on: `true` to set power on, `false` to set power off
     */
    public func power(on: Bool) {
        if on {
            if self.power {
                return
            }
            self.startTime = Int64(Date().timeIntervalSince1970)
            self.power = true
            reset()
        } else {
            self.power = false
        }
    }
    
    /**
     * Get whether or not the sensor is powered on.
     * - returns: `true`, if the sensor is powered on.
     */
    public func isPoweredOn() -> Bool {
        return self.power
    }
    
    /**
     * An action to reset the miniumum and maximum measured temperature to
     * the current value.
     */
    public func reset() {
        minTemp = .greatestFiniteMagnitude
        maxTemp = .leastNormalMagnitude
    }

    /**
     * Get the current temperature value.
     * The simulated readings will fluctuate +/- 2 degrees.
     *
     * - returns: temperature value
     */
    public func getTemp() -> Double {
        return temperatureQueue.sync(execute: {
            let amplitude:Double = (Double(maxThreshold) - Double(minThreshold))/2.0
            let angleinradians:Measurement =
                Measurement(value: Double(self.angle), unit: UnitAngle.degrees)
                    .converted(to: .radians)
            let delta:Double = amplitude * sin(angleinradians.value)
            var temp:Double = 0.0
            if self.angle == 90 {
                // violate max threshold when angle is 90 degrees
                temp = roundToNearest10th(d:Double(maxThreshold) + 2.0)
            } else if angle == 180 {
                // violate min threshold when angle is 180 degrees
                temp = roundToNearest10th(d:Double(minThreshold) - 2.0)
            } else {
                temp = roundToNearest10th(d:Double(minThreshold) + amplitude + delta)
            }
            self.angle = (angle + 15) % 360
            temp = clamp(val:temp, minimum:LOWER_TEMP_LIMIT, maximum:UPPER_TEMP_LIMIT)
            if temp < minTemp {
                minTemp = temp
            }
            if maxTemp < temp {
                maxTemp = temp
            }
            return temp
        })
    }
    
    /**
     * Round to nearest 10th.
     *
     * - parameter d: the value to round
     * - returns: the parameter value, rounded to the nearest 10th
     */
    func roundToNearest10th(d:Double) -> Double {
    
        if d == 0 {
            return d
        }
        
        let r:Double = (d < 0.0) ? -0.05 : 0.05
        let wholepart:Int = ((d + r) * 10 as NSNumber).intValue
        return (wholepart as NSNumber).doubleValue / 10.0
    }
    
    /**
     * Get the temperature units (scale)
     * - returns: the temperature units
     */
    public func getUnit() -> String {
        return "\u{00B0}C" //"Celsius"
    }
    
    /**
     * Get the minimum temperature since the last power on or reset.
     * - returns: the minimum temperature since the last power on or reset
     */
    public func getMinTemp() -> Double {
        return self.minTemp
    }
    
    /**
     * Get the maximum temperature since the last power on or reset.
     * - returns: the maximum temperature since the last power on or reset
     */
    public func getMaxTemp() -> Double {
        return self.maxTemp
    }
    
    /**
     * Get the minumum threshold value.
     * - returns: the minimum threshold
     */
    public func getMinThreshold() -> Int {
        return self.minThreshold
    }
    
    /**
     * Set the minimum degrees Celsius threshold for alerts. The value is
     * clamped to the range [-20..0].
     * - parameter threshold: a value between -20 and 0
     */
    public func setMinThreshold(threshold: Int) {
        if threshold < -20 {
            minThreshold = -20
        } else if threshold > 0 {
            minThreshold = 0
        } else {
            minThreshold = threshold
        }
    }
    
    /**
     * Get the maximum threshold value.
     * - returns: the maximum threshold
     */
    public func getMaxThreshold() -> Int {
        return self.maxThreshold
    }
    
    /**
     * Set the maximum degrees Celsius threshold for alerts. The value is
     * clamped to the range [65..80].
     * - parameter threshold: a value between 65 and 80
     */
    public func setMaxThreshold(threshold: Int) {
        if threshold < 65 {
            maxThreshold = 65
        } else if threshold > 80 {
            maxThreshold = 80
        } else {
            maxThreshold = threshold
        }
    }
    
    /**
     * Get the time at which the device was powered on or reset.
     * - returns: the device start time
     */
    public func getStartTime() -> Int64 {
        return self.startTime
    }
    /**
     * Get the manufacturer name, which can be used as part of the
     * device meta-data.
     * - returns the manufacturer name
     */
    public func getManufacturer() -> String {
        return "Sample"
    }
    
    /**
     * Get the model number, which can be used as part of the
     * device meta-data.
     * - returns: the model number
     */
    public func getModelNumber() -> String {
        return "MN-" + self.hardwareId
    }
    
    /**
     * Get the serial number, which can be used as part of the
     * device meta-data.
     * - returns: the serial number
     */
    public func getSerialNumber() -> String {
        return "SN-" + self.hardwareId
    }
    
    /**
     * Get the hardware id, which can be used as part of the
     * device meta-data.
     * - returns: the hardware id
     */
    public func getHardwareId() -> String {
        return self.hardwareId
    }
    /**
     * Clamp value between min and max, inclusive.
     * @param val the value
     * @param min the minimum allowable value
     * @param max the maximum allowable value
     * @return A value such that {@code min <= value <= max}
     */
    private func clamp(val:Double, minimum:Double, maximum:Double) -> Double {
        if val < minimum {
            return minimum
        }
        if val > maximum {
            return maximum
        }
        return val
    }
}
